import pddlgym
import imageio
import numpy as np
import torch
import numpy.random as npr
import random
import argparse

tiles = ['t1', 't2', 't3', 't4', 't5', 't6', 't7', 't8']
positions = ['x1', 'x2', 'x3', 'y1', 'y2', 'y3']
action_order = [f'(move-up {tile} {pos1} {pos2} {pos3})' for pos1 in positions for pos2 in positions for pos3 in positions for tile in tiles
                if pos1!=pos2 and pos1!=pos3 and pos2!=pos3]
action_order.extend([f'(move-down {tile} {pos1} {pos2} {pos3})' for pos1 in positions for pos2 in positions for pos3 in positions  for tile in tiles
                if pos1!=pos2 and pos1!=pos3 and pos2!=pos3])
action_order.extend([f'(move-left {tile} {pos1} {pos2} {pos3})' for pos1 in positions for pos2 in positions for pos3 in positions for tile in tiles
                if pos1!=pos2 and pos1!=pos3 and pos2!=pos3])
action_order.extend([f'(move-right {tile} {pos1} {pos2} {pos3})' for pos1 in positions for pos2 in positions for pos3 in positions for tile in tiles
                if pos1!=pos2 and pos1!=pos3 and pos2!=pos3])
action_order = {k: v for v, k in enumerate(action_order)}
propositions = [f'(at {tile} {pos1} {pos2})' for pos1 in positions for pos2 in positions if pos1!=pos2 for tile in tiles]
propositions.extend([f'(blank {pos1} {pos2})'  for pos1 in positions for pos2 in positions if pos1!=pos2])
propositions.extend([f'(inc {pos1} {pos2})'  for pos1 in positions for pos2 in positions if pos1!=pos2])
propositions.extend([f'(dec {pos1} {pos2})'  for pos1 in positions for pos2 in positions if pos1!=pos2])
propositions = {k: v for v, k in enumerate(propositions)}
print(propositions)

def obs_to_label(obs):
    label = np.zeros(len(propositions))
    for literal in obs.literals:
        if literal.pddl_str() not in propositions:
            # This is for handfull
            continue
        label[propositions[literal.pddl_str()]] = 1
    return label

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="-s save_addr")
    parser.add_argument("-s", default="data", help="save address")
    parser.add_argument("--seed", type=int, default=8800)
    args = parser.parse_args()

    torch.manual_seed(args.seed)
    random.seed(args.seed)
    npr.seed(args.seed)

    # Use PDDL Gym
    # https://github.com/tomsilver/pddlgym
    env = pddlgym.make("PDDLEnvSlidetile-v0", operators_as_actions=True, dynamic_action_space=True)
    obs, info = env.reset()

    labels = []
    actions = []

    for i in range(50): #5500
        img = env.render()
        imageio.imsave(f"{args.s}/{i}.png", img)
        del img
        labels.append(obs_to_label(obs))
        action = env.action_space.sample(obs)
        actions.append(action_order[action.pddl_str()])
        obs = env.step(action)[0]

    labels_tensor = torch.tensor(np.array(labels), dtype=torch.float32)
    actions_tensor = torch.tensor(np.array(actions))

    torch.save(labels_tensor, f"{args.s}/labels.pt")
    torch.save(actions_tensor, f"{args.s}/actions.pt")


#!/usr/bin/env python

import argparse
from itertools import product
import random
from collections import Counter
import sys


def print_solver(instance_name, width, height, s):
    print(width, height)
    for (x, y, n) in s:
        print(x, y, n)


def adjacent(width, height, x, y):
    res = []
    if x < width - 1:
        res.append((x + 1, y))
    if y < height - 1:
        res.append((x, y + 1))
    if x > 0:
        res.append((x - 1, y))
    if y > 0:
        res.append((x, y - 1))
    return res


def get_adjacent_pairs(width, height):
    res = []
    for x in range(0, width):
        for y in range(0, height):
            if x < width - 1:
                res.append((x, y, x + 1, y))
            if y < height - 1:
                res.append((x, y, x, y + 1))
            if x > 0:
                res.append((x, y, x - 1, y))
            if y > 0:
                res.append((x, y, x, y - 1))
    return res


def print_pddl(instance_name, width, height, s):
    max_number = max(n for (x, y, n) in s)
    number_objects = " ".join(["n%d" % n for n in range(1, max_number + 1)])

    positions = list(product(range(width), range(height)))
    available_positions = list(product(range(width), range(height)))
    blocked_positions = []
    partof_positions = []
    for i in range(len(s)):
        (x, y, n) = s[i]
        available_positions.remove((x, y))
        adj_list = adjacent(width, height, x, y)
        partof_positions += [(x, y, i) for (x, y) in adj_list]
        for p in adj_list:
            if p in available_positions:
                available_positions.remove(p)
            elif p not in blocked_positions:
                blocked_positions.append(p)

    partof_positions = [
        (x, y, z) for (x, y, z) in partof_positions if (x, y) not in blocked_positions
    ]

    # print available_positions
    # print blocked_positions
    # print partof_positions

    pos_objects = " ".join(["pos-%s-%s" % pos for pos in positions])

    group_objects = " ".join(["g%d" % n for n in range(0, len(s))])

    goal_facts = "\n ".join(["(group-painted g%d)" % n for n in range(0, len(s))])

    next_facts = "\n ".join(
        ["(NEXT n%d n%d)" % (n, n + 1) for n in range(0, max_number)]
    )

    is_adjacent_facts = "\n ".join(
        [
            "(CONNECTED pos-%s-%s pos-%s-%s)" % adj
            for adj in get_adjacent_pairs(width, height)
        ]
    )

    source_facts = "\n ".join(
        ["(SOURCE pos-%s-%s g%d)" % (s[i][0], s[i][1], i) for i in range(0, len(s))]
    )

    available_facts = "\n ".join(
        [f"(available pos-{x}-{y})" for (x, y) in available_positions]
    )

    blocked_facts = "\n ".join(
        [f"(blocked pos-{x}-{y})" for (x, y) in blocked_positions]
    )

    part_of_facts = "\n ".join(
        ["(part-of pos-%s-%s g%d)" % (x, y, i) for (x, y, i) in partof_positions]
    )

    remaining_cell_facts = "\n ".join(
        ["(remaining-cells g%d n%d)" % (i, s[i][2]) for i in range(0, len(s))]
    )

    print(
        """(define (problem {instance_name})
(:domain paint-nurikabe)
(:objects
  {pos_objects} - cell
  {number_objects} - num
  {group_objects} - group
)
(:init
 {next_facts}

 {is_adjacent_facts}

 (robot-pos pos-0-0)
 (moving)

 {source_facts}

 {available_facts}
 {blocked_facts}
 {part_of_facts}
 {remaining_cell_facts}
)
(:goal (and
 {goal_facts}
))
)
""".format(
            **locals()
        )
    )


def print_solution(width, height, m):
    print(width, height)
    for row in m:
        for cell in row:
            if cell == " ":
                cell = "."
            print(cell, end=" ")
        print


def is_number(x):
    try:
        d = int(x)
        return True
    except ValueError:
        return False


def is_empty(cell):
    return cell == " " or is_number(cell)


def reconstruct_islands(colored_map):
    result = [list(row) for row in colored_map]
    height = len(result)
    width = len(result[0])
    islands = []
    for y, row in enumerate(result):
        for x, cell in enumerate(row):
            if is_empty(cell):
                queue = [(x, y)]
                result[y][x] = "X"
                island_cells = []
                island_start = (x, y)
                while queue:
                    cx, cy = queue.pop()
                    island_cells.append((cx, cy))
                    for nx, ny in [
                        (cx + 1, cy),
                        (cx - 1, cy),
                        (cx, cy + 1),
                        (cx, cy - 1),
                    ]:
                        if 0 <= ny < height and 0 <= nx < width:
                            if is_number(result[ny][nx]):
                                island_start = (nx, ny)
                            if is_empty(result[ny][nx]):
                                queue.append((nx, ny))
                                result[ny][nx] = "X"
                islands.append((island_start, island_cells))
    for (sx, sy), cells in islands:
        n = len(cells)
        for x, y in cells:
            result[y][x] = " "
        result[sy][sx] = str(n)
    return result


def scale_down_instance(colored_map, target_width, target_height):
    scaled_map = [list(row[:target_width]) for row in colored_map[:target_height]]
    return reconstruct_islands(scaled_map)


def print_map(colored_map):
    for row in colored_map:
        print(" ".join(row))


def map_to_instance(colored_map):
    s = []
    for y, row in enumerate(colored_map):
        for x, cell in enumerate(row):
            if is_number(cell):
                s.append((x, y, int(cell)))
    return s


def random_map(width, height):
    STOPCHANCE = 0.01
    BRANCHCHANCE = 0.5
    cmap = [[" " for x in range(width)] for y in range(height)]
    ends = [(random.randint(0, width - 1), random.randint(0, height - 1))]
    cmap[ends[0][1]][ends[0][0]] = "#"
    next_ends = []
    while ends:
        for x, y in ends:
            if random.random() < STOPCHANCE:
                continue
            neighbors = []
            for nx, ny in [(x + 1, y), (x - 1, y), (x, y + 1), (x, y - 1)]:
                if 0 <= ny < height and 0 <= nx < width and cmap[ny][nx] == " ":
                    for nnx, nny in [
                        (nx + 1, ny),
                        (nx - 1, ny),
                        (nx, ny + 1),
                        (nx, ny - 1),
                    ]:
                        if (nnx, nny) == (x, y):
                            continue
                        if (
                            0 <= nny < height
                            and 0 <= nnx < width
                            and cmap[nny][nnx] == "#"
                        ):
                            break
                    else:
                        neighbors.append((nx, ny))
            if neighbors:
                successors = {random.choice(neighbors)}
                if random.random() < BRANCHCHANCE:
                    successors.add(random.choice(neighbors))
                for nx, ny in successors:
                    cmap[ny][nx] = "#"
                    next_ends.append((nx, ny))
        ends = next_ends
        next_ends = []
    return reconstruct_islands(cmap)


def reasonable_random_map(width, height):
    size = max(width, height)
    for i in range(1000):
        m = random_map(width, height)
        groups = Counter([s for x, y, s in map_to_instance(m)])
        if groups.get(1, 0) <= size / 2 and sorted(groups.items())[-1][0] < size:
            return m


def generate_state_lpo(args):
    url = "http://www.logicgamesonline.com/nurikabe/archive.php?pid=%d" % args.pid
    instance_name = "nurikabe-lpo%d-%dx%d" % (args.pid, args.width, args.height)
    from urllib.error import URLError
    from urllib.request import urlopen

    try:
        response = urlopen(url, timeout=5)
        content = response.read().decode("utf-8")
    except URLError as e:
        sys.exit(e)

    puzzle = ""
    for line in content.splitlines():
        if line.startswith('var solpuz = "'):
            puzzle = line[len('var solpuz = "') : -2]
            break
    if len(puzzle) != 9 * 9:
        sys.exit("Cannot find or parse puzzle")
    colored_map = [list(puzzle[i * 9 : (i + 1) * 9]) for i in range(9)]
    colored_map = scale_down_instance(colored_map, args.width, args.height)
    return instance_name, colored_map


def generate_state_random(args):
    instance_name = "random-%dx%d-%d" % (args.width, args.height, args.random_seed)
    colored_map = reasonable_random_map(args.width, args.height)
    return instance_name, colored_map


def parse():
    parser = argparse.ArgumentParser()
    parser.add_argument("source", choices=["lpo", "random"])
    # TODO: this should only be an option when using lpo
    parser.add_argument("pid", type=int, help="problem ID on logicgamesonline.com")
    parser.add_argument("width", type=int, help="Width of the board.")
    parser.add_argument("height", type=int, help="Height of the board.")
    parser.add_argument("random_seed", type=int)
    parser.add_argument("output_type", choices=["pddl", "raw", "solution"])
    return parser.parse_args()


def main():
    args = parse()
    random.seed(args.random_seed)

    if args.source == "lpo":
        instance_name, m = generate_state_lpo(args)
    elif args.source == "random":
        instance_name, m = generate_state_random(args)

    s = map_to_instance(m)

    if args.output_type == "raw":
        print_solver(instance_name, args.width, args.height, s)
    elif args.output_type == "pddl":
        print_pddl(instance_name, args.width, args.height, s)
    elif args.output_type == "solution":
        print_solution(args.width, args.height, m)


if __name__ == "__main__":
    main()



from models.rosame import *
from models.cv_gridworld import *
from data.dataset import *

import torch
import torch.optim as optim

from torch.utils.data import Dataset, DataLoader, random_split
from torchvision import transforms
import torchvision

import random
import argparse
import os


def get_domain_model(domain, device):
    domain_model = load_model(
        os.path.join(
            os.path.dirname(__file__), "models/domains", domain, "domain_model.json"
        ),
        device,
    )
    domain_model.ground_from_json(
        os.path.join(
            os.path.dirname(__file__), "models/domains", domain, "objects.json"
        )
    )
    return domain_model


@torch.no_grad()
def compute_correctness(pred_flat, target_flat):
    """
    Expect input in the shape of (batch_size, trace_len, prop_num)
    """
    trace_len = pred_flat.shape[1]
    prop_num = pred_flat.shape[2]

    pred = pred_flat.reshape(-1, prop_num)
    target = target_flat.reshape(-1, prop_num)
    pred = (pred > 0.5).float()

    correct = torch.sum(torch.isclose(pred, target))

    return float(correct) / prop_num / trace_len


def run(
    epoch,
    cv_model,
    domain_model,
    optimizer,
    data_loader,
    gamma,
    lambda_,
    device,
    to_train=False,
):
    loss_final, acc_running, data_num = 0, 0, 0

    if to_train:
        cv_model.train()  # Set model to training mode
    else:
        cv_model.eval()  # Set model to evaluate mode

    for i, (data, label, action) in enumerate(data_loader):
        data = data.to(device)
        label = label.to(device)
        action = action.to(device)
        trace_len = action.shape[1]
        flattened_data = data.flatten(start_dim=0, end_dim=1)

        with torch.set_grad_enabled(to_train):
            preds = cv_model(flattened_data)
            loss = 0
            # Domain model inference loss
            precon, addeff, deleff = domain_model.build(action.flatten())
            domain_preds = preds * (1 - deleff) + (1 - preds) * addeff
            # domain_preds = 1 - (1-preds*(1-deleff)) * (1-(1-preds)*addeff)
            validity_constraint = (1 - preds) * (precon)
            preds = preds.unflatten(0, (-1, trace_len))
            domain_preds = domain_preds.unflatten(0, (-1, trace_len))
            loss += F.mse_loss(domain_preds[:, :-1], preds[:, 1:], reduction="sum")
            loss += gamma * F.mse_loss(
                domain_preds[:, -1], label[:, -1], reduction="sum"
            )
            loss += F.mse_loss(
                validity_constraint,
                torch.zeros(
                    validity_constraint.shape,
                    dtype=validity_constraint.dtype,
                    device=device,
                ),
                reduction="sum",
            )
            # Add a prior
            loss += lambda_ * F.mse_loss(
                precon,
                torch.ones(precon.shape, dtype=precon.dtype, device=device),
                reduction="sum",
            )
            if to_train:
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

        loss_final += loss.item()
        acc_running += compute_correctness(preds.data, label[:, :-1])
        data_num += preds.data.shape[0]

    if to_train:
        print(
            "Epoch {} TRAINING SET RESULTS: Average loss: {:.4f} Acc: {:.4f}".format(
                epoch, loss_final, acc_running / data_num
            )
        )
    else:
        print(
            "Epoch {} TESTING SET RESULTS: Average loss: {:.4f} Acc: {:.4f}".format(
                epoch, loss_final, acc_running / data_num
            )
        )

    torch.cuda.empty_cache()


if __name__ == "__main__":
    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--domain",
        choices=[
            "grid_blocks",
            "grid_gripper",
            "grid_logistics",
            "synth_blocks",
            "synth_hanoi",
            "synth_8-puzzle",
        ],
    )
    parser.add_argument("--gamma", type=float, default=10)
    parser.add_argument("--lambda_", type=float, default=0.2)
    parser.add_argument("--epochs", type=int, default=100)
    parser.add_argument("--lr_schema", type=float, default=1e-3)
    parser.add_argument("--lr_gridcv_grid", type=float, default=1e-5)
    parser.add_argument("--lr_gridcv_mlp", type=float, default=1e-3)
    parser.add_argument("--lr_synthcv", type=float, default=1e-3)
    parser.add_argument("--batch_size", type=int, default=128)
    parser.add_argument("--dataset_pth")
    parser.add_argument("--trace_num", type=int)
    parser.add_argument("--trace_len", type=int)
    parser.add_argument("--block_num", type=int, default=5)
    parser.add_argument("--ball_num", type=int, default=6)
    parser.add_argument("--seed", type=int, default=8800)
    parser.add_argument("-s", default="model.pth", help="save address")
    args = parser.parse_args()

    random.seed(args.seed)
    torch.manual_seed(args.seed)

    # Set up domain model and cv model.
    # Gather experiment data.
    if args.domain == "grid_blocks":
        block_num = args.block_num
        domain_model = get_domain_model("blocks", device)
        cv_model = CVGrid(
            GridConv(digit_class_num=block_num + 1, input_channel=1),
            block_dim=(block_num + 1, block_num),
            block_size=28,  # MNIST images are 28x28
            hidden_dim=128,
            digit_class_num=block_num + 1,
            prop_dim=len(domain_model.propositions),
        )
        data_transform = RearrangeColumn(block_num)
    elif args.domain == "grid_gripper":
        ball_num = args.ball_num
        domain_model = get_domain_model("gripper", device)
        cv_model = CVGrid(
            GridConv(digit_class_num=(ball_num + 1) * 2, input_channel=1),
            block_dim=(4, ball_num),
            block_size=28,  # MNIST images are 28x28
            hidden_dim=128,
            digit_class_num=(ball_num + 1) * 2,
            prop_dim=len(domain_model.propositions),
        )
        data_transform = RearrangeBalls(ball_num)
    elif args.domain == "grid_logistics":
        domain_model = get_domain_model("logistics", device)
        digit_class_num = 35
        cv_model = CVGrid(
            GridConv(digit_class_num=digit_class_num, input_channel=3),
            block_dim=(6, 6),
            block_size=28,  # MNIST images are 28x28
            hidden_dim=256,
            digit_class_num=digit_class_num,
            prop_dim=len(domain_model.propositions),
        )
        data_transform = RearrangeItems()
    elif args.domain == "synth_blocks":
        domain_model = get_domain_model("blocks", device)
        cv_model = torchvision.models.resnet18()
        cv_model.fc = nn.Sequential(
            nn.Linear(512, 512),
            nn.ReLU(),
            nn.Linear(512, 256),
            nn.ReLU(),
            nn.Linear(256, len(domain_model.propositions)),
        )
        data_transform = transforms.Compose(
            [
                transforms.Resize(64),
                transforms.RandomHorizontalFlip(0.5),
            ]
        )
    elif args.domain == "synth_hanoi":
        domain_model = get_domain_model("hanoi", device)
        cv_model = torchvision.models.resnet18()
        cv_model.fc = nn.Sequential(
            nn.Linear(512, 512),
            nn.ReLU(),
            nn.Linear(512, 256),
            nn.ReLU(),
            nn.Linear(256, len(domain_model.propositions)),
        )
        data_transform = transforms.Resize(64)
    elif args.domain == "synth_8-puzzle":
        domain_model = get_domain_model("8-puzzle", device)
        cv_model = torchvision.models.resnet18()
        cv_model.fc = nn.Sequential(
            nn.Linear(512, 512),
            nn.ReLU(),
            nn.Linear(512, 256),
            nn.ReLU(),
            nn.Linear(256, len(domain_model.propositions)),
        )
        data_transform = transforms.Resize(64)

    domain_model = domain_model.to(device)
    cv_model = cv_model.to(device)

    # Get Dataset
    if args.domain.startswith("grid"):
        dataset = GridDataset(
            args.dataset_pth, args.trace_len, transforms=data_transform
        )
        training_size = int(args.trace_num * 0.9)
        trainset, testset, _ = random_split(
            dataset,
            [
                training_size,
                args.trace_num - training_size,
                len(dataset) - args.trace_num,
            ],
        )
    else:
        skip = "break_symmetry" if args.domain == "synth_blocks" else 1
        dataset = SynthDataset(
            args.dataset_pth, args.trace_len, skip, transforms=data_transform
        )
        trainset, testset, _ = random_split(
            dataset, [args.trace_num, 100, len(dataset) - args.trace_num - 100]
        )
    train_loader = DataLoader(trainset, args.batch_size, shuffle=True)
    test_loader = DataLoader(testset, args.batch_size, shuffle=True)

    # Create optimizer
    parameters = []
    for schema in domain_model.action_schemas:
        parameters.append({"params": schema.parameters(), "lr": args.lr_schema})
    if args.domain.startswith("grid"):
        parameters.extend(
            [
                {"params": cv_model.mlp.parameters(), "lr": args.lr_gridcv_mlp},
                {
                    "params": cv_model.grid_convnet.parameters(),
                    "lr": args.lr_gridcv_grid,
                },
            ]
        )
    else:
        parameters.extend([{"params": cv_model.parameters(), "lr": args.lr_synthcv}])
    optimizer = optim.Adam(parameters)

    print("---------------------------------")
    print("Domain:", args.domain)
    print("Gamma:", args.gamma)
    print("Lambda:", args.lambda_)
    print("Trace Num:", args.trace_num)
    print("Trace Len:", args.trace_len)
    for epoch in range(args.epochs):
        run(
            epoch,
            cv_model,
            domain_model,
            optimizer,
            train_loader,
            args.gamma,
            args.lambda_,
            device,
            True,
        )
        run(
            epoch,
            cv_model,
            domain_model,
            optimizer,
            test_loader,
            args.gamma,
            args.lambda_,
            device,
            False,
        )
    for schema in domain_model.action_schemas:
        schema.pretty_print()
        print()

    torch.save(
        {"domain_model": domain_model.state_dict(), "cv_model": cv_model.state_dict()},
        args.s,
    )



import numpy as np
import numpy.random as npr
import matplotlib.pyplot as plt
import torch
import torchvision.datasets as datasets
from macq import generate, extract
import argparse
import os

mnist_dataset, target_index = None, None


def init_mnist():
    global mnist_dataset, target_index
    mnist_dataset = datasets.MNIST(root='./data', train=True, download=True, transform=None)
    targets = mnist_dataset.targets
    target_index = {}
    for i in range(10):
        target_index[i] = np.argwhere(targets==i).flatten().tolist()


def state_to_img(state):
    # Turn a macq state to an image
    img = np.empty((grid_shape[0], grid_shape[1], 28, 28))
    
    room_available_loc = [list(range(num_balls)) for _ in range(len(rooms))]
    free_ball_loc = {}
    backgrounds = {(i, j) for i in range(grid_shape[0]) for j in range(grid_shape[1])}
    
    for i in range(len(rooms)):
        if state.fluents[all_fluents[f'(at-robby room {rooms[i]})']]:
            gripper_loc = 2*i
    
    for i in range(1, num_balls+1):
        ball_free = False
        for j in range(len(rooms)):
            if state.fluents[all_fluents[f'(at ball ball{i} room {rooms[j]})']]:
                loc = npr.choice(room_available_loc[j])
                img[2*j+1, loc] = np.asarray(mnist_dataset.__getitem__(npr.choice(target_index[i]))[0])
                free_ball_loc[f'ball{i}'] = (2*j+1, loc)
                room_available_loc[j].remove(loc)
                # Remove the location from background
                backgrounds.remove((2*j+1, loc))
                ball_free = True
                break
        
        if ball_free:
            continue
            
        for g in range(len(grippers)):
            if state.fluents[all_fluents[f'(carry ball ball{i} gripper {grippers[g]})']]:
                img[gripper_loc, g] = 255-np.asarray(mnist_dataset.__getitem__(npr.choice(target_index[i]))[0])
                # Remove the location from background
                backgrounds.remove((gripper_loc, g))
                break
                
    for g in range(len(grippers)):
        if state.fluents[all_fluents[f'(free gripper {grippers[g]})']]:
            img[gripper_loc, g] = 255-np.asarray(mnist_dataset.__getitem__(npr.choice(target_index[0]))[0])
            # Remove the location from background
            backgrounds.remove((gripper_loc, g))
            
    for i, j in backgrounds:
        img[i, j] = np.asarray(mnist_dataset.__getitem__(npr.choice(target_index[0]))[0])
        
    return img, (free_ball_loc, room_available_loc)


def get_next_img(current_img, meta, action):
    next_img = current_img.copy()
    free_ball_loc, room_available_loc = meta
    
    if action.name=='move':
        room_from = rooms.index(action.obj_params[0].name)
        room_to = rooms.index(action.obj_params[1].name)
        temp = next_img[room_from*2, list(range(len(grippers)))]
        next_img[room_from*2, list(range(len(grippers)))] = next_img[room_to*2, list(range(len(grippers)))]
        next_img[room_to*2, list(range(len(grippers)))] = temp
    elif action.name=='pick':
        room = rooms.index(action.obj_params[1].name)
        ball_loc = free_ball_loc[action.obj_params[0].name]
        room_loc = room*2
        gripper_loc = grippers.index(action.obj_params[2].name)
        temp = next_img[room_loc, gripper_loc].copy()
        next_img[room_loc, gripper_loc] = 255-next_img[ball_loc]
        next_img[ball_loc] = 255-temp
        del free_ball_loc[action.obj_params[0].name]
        room_available_loc[room].append(ball_loc[1])
    elif action.name=='drop':
        room = rooms.index(action.obj_params[1].name)
        ball_loc = (room*2+1, npr.choice(room_available_loc[room]))
        room_loc = room*2
        gripper_loc = grippers.index(action.obj_params[2].name)
        temp = next_img[room_loc, gripper_loc].copy()
        next_img[room_loc, gripper_loc] = 255-next_img[ball_loc]
        next_img[ball_loc] = 255-temp
        free_ball_loc[action.obj_params[0].name] = ball_loc
        room_available_loc[room].remove(ball_loc[1])
        
    return next_img, meta


def state_to_label(state):
    label = np.zeros(len(fluents))
    for f in state.fluents:
        label[fluents[f._serialize()]] = state.fluents[f]
    return label


def show_mnist_image(raw):
    plt.figure()
    digits = np.concatenate(np.concatenate(raw,axis=1), axis=1).astype(np.uint8)

    #img = Image.fromarray(255-board)
    plt.imshow(255-digits, cmap='gray')


def get_actions_and_props(balls, rooms, grippers):
    actions = [f'move room {room_from} room {room_to}'
               for room_from in rooms
               for room_to in rooms
               if room_from!=room_to]
    actions.extend([f'pick ball {ball} gripper {gripper} room {room}'
                    for ball in balls for gripper in grippers for room in rooms])
    actions.extend([f'drop ball {ball} gripper {gripper} room {room}'
                    for ball in balls for gripper in grippers for room in rooms])
    actions = {k: v for v, k in enumerate(actions)}

    fluents = [f'(at-robby room {room})' for room in rooms]
    fluents.extend([f'(at ball {ball} room {room})' for ball in balls for room in rooms])
    fluents.extend([f'(free gripper {gripper})' for gripper in grippers])
    fluents.extend([f'(carry ball {ball} gripper {gripper})' for ball in balls for gripper in grippers])
    fluents = {k: v for v, k in enumerate(fluents)}
    return actions, fluents



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-b", type=int, default=6, help="ball num")
    parser.add_argument("--rooms", action='store', type=str, nargs='*', default=["rooma", "roomb"])
    parser.add_argument("--grippers", action='store', type=str, nargs='*', default=["left", "right"])
    parser.add_argument("-t", type=int, default=1000, help="trace num")
    parser.add_argument("-l", type=int, default=5, help="trace length")
    parser.add_argument("--skip", type=int, default=1, help="skip between traces")
    parser.add_argument("-s", default="data", help="save address")
    parser.add_argument("--pddl_dom", default="/Users/omarwattad/Documents/Action Model - Research/rosame/data/pddl/gripper/domain2.pddl")
    parser.add_argument("--pddl_prob", default="/Users/omarwattad/Documents/Action Model - Research/rosame/data/pddl/gripper/prob01.pddl")
    parser.add_argument("--seed", type=int, default=8800)
    args = parser.parse_args()

    torch.manual_seed(args.seed)
    npr.seed(args.seed)

    # Grid shape: 2*room_num, ball_num(>=2)
    grid_shape = (2*len(args.rooms), args.b)
    num_balls = args.b
    balls = [f'ball{i}' for i in range(1, num_balls+1)]
    rooms = args.rooms
    grippers = args.grippers
    save_to = args.s
    init_mnist()
    actions, fluents = get_actions_and_props(balls, rooms, grippers)

    trace = generate.pddl.VanillaSampling(dom=args.pddl_dom,
                                          prob=args.pddl_prob,
                                          plan_len = args.t*(args.l+args.skip), num_traces = 1).traces[0]
    all_fluents = {f._serialize():f for f in trace.fluents}
    traces_images = []
    traces_actions = []
    traces_labels = []

    for n in range(args.t):
        trace_images = []
        trace_actions = []
        trace_labels = []
        img, meta = state_to_img(trace.steps[n*(args.l+args.skip)].state)
        for step in range(n*(args.l+args.skip), n*(args.l+args.skip)+args.l):
            trace_images.append(img)
            trace_labels.append(state_to_label(trace.steps[step].state))
            # action parameter order may not be the same as our model
            action = trace.steps[step].action
            action_obj_params = sorted([o for o in action.obj_params], key=lambda o:o.obj_type)
            trace_actions.append(actions[f"{action.name} {' '.join([o.details()for o in action_obj_params])}"])
            img, meta = get_next_img(img, meta, trace.steps[step].action)

            # to debug: if traj. still not valid, implement and run this: val(domain, pre_state, action, next_state) --> true if the transition is valid (precond. satisfied, effects also)

        # Record final state label
        trace_labels.append(state_to_label(trace.steps[n*(args.l+args.skip)+args.l].state))
        
        traces_images.append(trace_images)
        traces_actions.append(trace_actions)
        traces_labels.append(trace_labels)

    traces_images_tensor = torch.tensor(np.array(traces_images))
    traces_actions_tensor = torch.tensor(np.array(traces_actions))
    traces_labels_tensor = torch.tensor(np.array(traces_labels), dtype=torch.float32)

    torch.save(traces_images_tensor, os.path.join(save_to, "features_img.pt"))
    torch.save(traces_actions_tensor, os.path.join(save_to, "actions.pt"))
    torch.save(traces_labels_tensor, os.path.join(save_to, "labels.pt"))


#!/usr/bin/env python3

import argparse
from itertools import product
import random
import itertools

import sys


def print_pddl(instance_name, num_floors, elevators, passengers, cost_slow, cost_fast):

    num_counts = max(num_floors, max(e.capacity + 1 for e in elevators))

    count_objects = [f"n{i}" for i in range(num_counts)]
    next_facts = "\n       ".join(
        [
            f"(next {count_objects[i]} {count_objects[i+1]})"
            for i in range(num_floors - 1)
        ]
    )
    above_facts = "\n       ".join(
        [
            f"(above {count_objects[i]} {count_objects[j]})"
            for i in range(num_floors - 1)
            for j in range(i + 1, num_floors)
        ]
    )

    elevators_facts = "\n       ".join([e.get_init() for e in elevators])
    passenger_facts = "\n       ".join([p.get_init() for p in passengers])
    goal_facts = "\n       ".join([p.get_goal() for p in passengers])

    passenger_objects = " ".join([p.name for p in passengers])
    fast_elevator_objects = " ".join(
        [e.name for e in elevators if e.name.startswith("fast")]
    )
    slow_elevator_objects = " ".join(
        [e.name for e in elevators if e.name.startswith("slow")]
    )

    slow_floors = set()
    fast_floors = set()
    for e in elevators:
        if e.name.startswith("slow"):
            slow_floors.update(itertools.combinations(e.floors, 2))
        else:
            fast_floors.update(itertools.combinations(e.floors, 2))

    cost_facts = "\n       ".join(
        [
            f"(= (travel-slow n{i} n{j}) {cost_slow.get_cost(i,j)})"
            for i, j in sorted(slow_floors)
        ]
        + [
            f"(= (travel-fast n{i} n{j}) {cost_fast.get_cost(i,j)})"
            for i, j in sorted(fast_floors)
        ]
    )

    print(
        f"""(define (problem {instance_name})
(:domain elevators-sequencedstrips)
    (:objects
       {" ".join(count_objects)} - count
       {passenger_objects} - passenger
       {fast_elevator_objects} - fast-elevator
       {slow_elevator_objects} - slow-elevator
    )

    (:init
       {next_facts}

       {above_facts}

       {elevators_facts}

       {passenger_facts}

       {cost_facts}

       (= (total-cost) 0)
    )

    (:goal (and
       {goal_facts}
    ))

    (:metric minimize (total-cost))
)
    """
    )


class Elevator:
    def __init__(self, name, capacity, first_floor, last_floor, step_size):

        self.name = name
        self.capacity = capacity
        self.floors = range(first_floor, last_floor, step_size)

        self.init_floor = random.choice(self.floors)

    def get_init(self):
        can_hold = " ".join(
            [f"(can-hold {self.name} n{i+1})" for i in range(self.capacity)]
        )
        reachable_floor = " ".join(
            [f"(reachable-floor {self.name} n{i})" for i in self.floors]
        )

        return f"""(lift-at {self.name} n{self.init_floor})
       (passengers {self.name} n0)
       {can_hold}
       {reachable_floor}
    """


class Passenger:
    def __init__(self, identifier, num_floors):
        self.name = f"p{identifier}"
        self.init_floor = random.randint(0, num_floors - 1)
        self.dest_floor = random.randint(0, num_floors - 1)
        while self.init_floor == self.dest_floor:
            self.dest_floor = random.randint(0, num_floors - 1)

    def get_init(self):
        return f"(passenger-at {self.name} n{self.init_floor})"

    def get_goal(self):
        return f"(passenger-at {self.name} n{self.dest_floor})"


class Cost:
    def __init__(self, cost_stop, cost_floor):
        self.cost_stop = cost_stop
        self.cost_floor = cost_floor

    def get_cost(self, i, j):
        return self.cost_stop + abs(i - j) * self.cost_floor


def parse():
    parser = argparse.ArgumentParser()

    parser.add_argument("areas", type=int, help="how many areas")
    parser.add_argument("area_size", type=int, help="")

    parser.add_argument("passengers", type=int, help="how many floors")

    parser.add_argument("fast_elevators", type=int, help="")
    parser.add_argument("slow_elevators", type=int, help="")

    parser.add_argument("--seed", type=int, default=1)

    parser.add_argument("--fast_capacity", type=int, default=6, help="")
    parser.add_argument("--slow_capacity", type=int, default=4, help="")
    parser.add_argument("--slow_cost", type=int, default=1, help="")
    parser.add_argument("--stop_slow_cost", type=int, default=5, help="")
    parser.add_argument("--fast_cost", type=int, default=3, help="")
    parser.add_argument("--stop_fast_cost", type=int, default=1, help="")

    parser.add_argument("--output", default=None)

    return parser.parse_args()


def main():
    args = parse()
    random.seed(args.seed)

    assert args.area_size >= 2

    floors = args.areas * args.area_size + 1
    fast_step_size = max(1, args.area_size // 2)
    slow_step_size = 1

    elevators = [
        Elevator(f"fast{i}", args.fast_capacity, 0, floors, fast_step_size)
        for i in range(args.fast_elevators)
    ] + [
        Elevator(
            f"slow{j}-{i}",
            args.slow_capacity,
            j * args.area_size,
            min(floors, (j + 1) * (args.area_size) + 1),
            slow_step_size,
        )
        for i in range(args.slow_elevators)
        for j in range(args.areas)
    ]

    passengers = [Passenger(i, floors) for i in range(args.passengers)]

    instance_name = "elevators"

    slow_cost = Cost(args.stop_slow_cost, args.slow_cost)
    fast_cost = Cost(args.stop_fast_cost, args.fast_cost)

    if args.output:
        sys.stdout = open(f"{args.output}/{instance_name}.pddl", "w")

    print(
        f"""; {floors} floors, area size: {args.area_size}, areas: {args.areas}, passengers: {args.passengers}"""
    )

    print_pddl(instance_name, floors, elevators, passengers, slow_cost, fast_cost)


if __name__ == "__main__":
    main()



import numpy


# A ring of small machines 2-7 with a fast connection between themselves. A super computer
# (more memory and faster cpu) is in the middle but the connection cost is high.

FILE_SIZE_DISTRIBUTION = lambda: numpy.random.random_integers(1, 5)

SERVERS = {
    "server1": {
        "capacity": 16,
        "io": 1,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(10, 2)))),
    },
    "server2": {
        "capacity": 8,
        "io": 2,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(20, 4)))),
    },
    "server3": {
        "capacity": 8,
        "io": 2,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(20, 4)))),
    },
    "server4": {
        "capacity": 8,
        "io": 2,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(20, 4)))),
    },
    "server5": {
        "capacity": 8,
        "io": 2,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(20, 4)))),
    },
    "server6": {
        "capacity": 8,
        "io": 2,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(20, 4)))),
    },
    "server7": {
        "capacity": 8,
        "io": 2,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(20, 4)))),
    },
}


def union(dict1, dict2):
    return dict(list(dict1.items()) + list(dict2.items()))


CONNECTIONS = union(
    {(1, i): {"send": 20} for i in range(2, 8)},
    {(i, i + 1 if i < 7 else 2): {"send": 2} for i in range(2, 8)},
)



import torch
import torch.nn as nn
import torch.nn.functional as F


class GridConv(nn.Module):
    """
    Convolutional neural network for MNIST digit recognition. From:
    https://github.com/pytorch/examples/blob/master/mnist/main.py
    """

    def __init__(self, digit_class_num, input_channel=1):
        super(GridConv, self).__init__()
        self.conv1 = nn.Conv2d(input_channel, 20, 5, 1)
        self.conv2 = nn.Conv2d(20, 50, 5, 1)
        self.fc1 = nn.Linear(4 * 4 * 50, 500)
        self.fc2 = nn.Linear(500, digit_class_num)

    def forward(self, x):
        x = F.relu(self.conv1(x))
        x = F.max_pool2d(x, 2, 2)
        x = F.relu(self.conv2(x))
        x = F.max_pool2d(x, 2, 2)
        x = x.view(-1, 4 * 4 * 50)
        x = F.relu(self.fc1(x))
        x = self.fc2(x)
        return x


class CVGrid(nn.Module):
    def __init__(
        self, grid_convnet, block_dim, block_size, hidden_dim, digit_class_num, prop_dim
    ):
        super(CVGrid, self).__init__()

        self.block_dim = block_dim
        self.hidden_dim = hidden_dim
        self.digit_class_num = digit_class_num
        self.prop_dim = prop_dim

        self.grid_convnet = grid_convnet

        self.mlp = nn.Sequential(
            nn.Linear(
                self.block_dim[0] * self.block_dim[1] * self.digit_class_num,
                self.hidden_dim * 4,
            ),
            nn.ReLU(),
            nn.Linear(self.hidden_dim * 4, self.hidden_dim * 2),
            nn.ReLU(),
            nn.Linear(self.hidden_dim * 2, self.hidden_dim),
            nn.ReLU(),
            nn.Linear(self.hidden_dim, self.prop_dim),
        )

    def forward(self, x):
        pred = x.flatten(start_dim=0, end_dim=2)
        pred = self.grid_convnet(pred)
        pred = pred.view(
            -1, self.block_dim[0] * self.block_dim[1] * self.digit_class_num
        )
        pred = self.mlp(pred)

        return F.sigmoid(pred)



#! /usr/bin/env python

import sys
import random


def help():
    print("usage: generator.py <n_couples> <n_cars> <n_places> [<seed>]")
    print("\t for solvability, cars should be at list n_couples + 1.")
    sys.exit(2)


if len(sys.argv) not in [4, 5]:
    help()

couples = int(sys.argv[1])
cars = int(sys.argv[2])
places = int(sys.argv[3])

if len(sys.argv) == 5:
    seed = int(sys.argv[4])
else:
    seed = None

random.seed(seed)

# hey, everything is ok. Let's start to do something.

print(
    "(define (problem Hiking-"
    + str(couples)
    + "-"
    + str(cars)
    + "-"
    + str(places)
    + ")"
)
print("(:domain hiking)")
print("(:objects ")

for i in range(cars):
    sys.stdout.write(" car" + str(i))

print(" - car")

for i in range(couples):
    sys.stdout.write(" tent" + str(i))

print(" - tent")

for i in range(couples):
    sys.stdout.write(" couple" + str(i))

print(" - couple")

for i in range(places):
    sys.stdout.write(" place" + str(i))

print(" - place")

for i in range(couples):
    sys.stdout.write(" guy" + str(i) + " girl" + str(i))

print(" - person")

print(")\n(:init")

for i in range(couples):
    print("(partners couple" + str(i) + " guy" + str(i) + " girl" + str(i) + ")")
    print("(at_person guy" + str(i) + " place0)")
    print("(at_person girl" + str(i) + " place0)")
    print("(walked couple" + str(i) + " place0)")
    print("(at_tent tent" + str(i) + " place0)")

    # random
    if random.randint(1, 2) == 2:
        print("(up tent" + str(i) + ")")
    else:
        print("(down tent" + str(i) + ")")

for i in range(cars):
    print("(at_car car" + str(i) + " place0)")

for i in range(places - 1):
    print("(next place" + str(i) + " place" + str(i + 1) + ")")

print(")\n(:goal\n(and")

for i in range(couples):
    print("(walked couple" + str(i) + " place" + str(places - 1) + ")")

print(")\n)\n)")



#!/usr/bin/env python

# This generator is very similar to the one used in IPC'18 for the TERMES domain. The main differences are:
#
# 1) We restrict ourselves to problems that go from an empty to a random board (in the IPC
#    all selected instances started in an empty board, but some instances were based on
#    manually-designed boards for the goal).
#
# 2) We add an extra parameter to control the instances generated: number of blocks.
#
# 3) We modify the output of a plan, so that it is included in the problem file.


import argparse
import random
import numpy
from solve import solve

import gen_random_tower_boards


def generate_readable(
    instance_name, grid_size_x, grid_size_y, max_height, robots, depots, i_grid, g_grid
):
    assert len(i_grid) == grid_size_y
    assert len(i_grid[0]) == grid_size_x
    lines = [instance_name]
    lines += ["Initial state:"]
    for y in range(grid_size_y):
        cells = []
        for x in range(grid_size_x):
            r = "R" if (x, y) in robots else " "
            h = str(i_grid[y][x])
            d = "D" if (x, y) in depots else " "
            cells.append(r + h + d)
        lines += [" ".join(cells)]
    lines += ["Goal state:"]
    for y in range(grid_size_y):
        cells = []
        for x in range(grid_size_x):
            if g_grid[y][x] is None:
                cells += [" * "]
            else:
                cells += [" %s " % g_grid[y][x]]
        lines += [" ".join(cells)]
    lines += ["Maximal height: %d" % max_height]
    return lines


def generate_for_pddl(
    instance_name,
    grid_size_x,
    grid_size_y,
    max_height,
    robots,
    depots,
    i_grid,
    g_grid,
    MULTI_ROBOT,
):
    comments = generate_readable(
        instance_name,
        grid_size_x,
        grid_size_y,
        max_height,
        robots,
        depots,
        i_grid,
        g_grid,
    )

    objects = []
    objects += ["n%d - numb" % (n) for n in range(max_height + 1)]
    objects += [
        "pos-%d-%d - position" % (x, y)
        for x in range(grid_size_x)
        for y in range(grid_size_y)
    ]

    facts = []
    facts += [
        "(height pos-%d-%d n%d)" % (x, y, i_grid[y][x])
        for x in range(grid_size_x)
        for y in range(grid_size_y)
    ]

    if MULTI_ROBOT:
        objects += ["r-%s - robot" % robot_id for robot_id in range(len(robots))]
        facts += ["(occupied pos-%d-%d)" % (x, y) for (x, y) in robots]
        facts += [
            "(at r-%d pos-%d-%d)" % (robot_id, x, y)
            for robot_id, (x, y) in enumerate(robots)
        ]
    else:
        assert len(robots) == 1
        facts += ["(at pos-%d-%d)" % (x, y) for (x, y) in robots]

    # Static predicates
    facts += ["(SUCC n%d n%d)" % (n + 1, n) for n in range(max_height)]
    facts += [
        "(NEIGHBOR pos-%d-%d pos-%d-%d)" % (x, y, x + dx, y + dy)
        for x in range(grid_size_x)
        for y in range(grid_size_y)
        for dx, dy in [(-1, 0), (1, 0), (0, -1), (0, 1)]
        if 0 <= x + dx < grid_size_x and 0 <= y + dy < grid_size_y
    ]
    facts += ["(IS-DEPOT pos-%d-%d)" % (x, y) for (x, y) in depots]

    goals = [
        "(height pos-%d-%d n%d)" % (x, y, g_grid[y][x])
        for x in range(grid_size_x)
        for y in range(grid_size_y)
        if g_grid[y][x] is not None
    ]
    goals += ["(not (has-block))"]

    lines = []
    lines += ["(define (problem %s)" % instance_name]
    lines += ["(:domain termes)"]
    lines += ["; %s" % c for c in comments]
    lines += ["(:objects"]
    lines += ["    %s" % o for o in objects]
    lines += [")", "(:init"]
    lines += ["    %s" % f for f in facts]
    lines += [")", "(:goal (and"]
    lines += ["    %s" % g for g in goals]
    lines += ["))"]
    lines += [")"]
    return lines


def remove_slack(grid):
    num_rows = len(grid)
    num_columns = len(grid[0])
    non_null_rows = sorted(
        {y for x in range(num_columns) for y in range(num_rows) if grid[y][x] > 0}
    )

    if not non_null_rows:
        return [[0]]

    grid = grid[non_null_rows[0] : non_null_rows[-1] + 1]

    num_rows = len(grid)
    num_columns = len(grid[0])

    non_null_columns = sorted(
        {x for x in range(num_columns) for y in range(num_rows) if grid[y][x] > 0}
    )

    grid = list(map(lambda x: x[non_null_columns[0] : non_null_columns[-1] + 1], grid))

    return grid


def get_value(grid, x, y, dx, dy):
    rx = x - dx
    ry = y - dy

    if ry < 0 or ry >= len(grid) or rx < 0 or rx >= len(grid[ry]):
        return 0

    return grid[ry][rx]


def parse():
    parser = argparse.ArgumentParser()

    parser.add_argument("--size_x", type=int, default=4)
    parser.add_argument("--size_y", type=int, default=4)
    parser.add_argument("--min_height", type=int, default=1)
    parser.add_argument("--max_height", type=int, default=4)
    parser.add_argument("--num_towers", type=int, default=4)

    parser.add_argument("--ensure_plan_tries", type=int, default=20)

    parser.add_argument("--ensure_plan", action="store_true")
    parser.add_argument("--dont_remove_slack", action="store_true")
    parser.add_argument("--store_plan", action="store_true")

    parser.add_argument("random_seed", type=int)
    parser.add_argument("output_type", choices=["pddl", "pddlfile", "readable"])
    return parser.parse_args()


def place_depots(grid_size_x, grid_size_y, init_grid, goal_grid):
    candidate_cells = [
        (y, abs(x - grid_size_x / 2), x, y)
        for x in range(grid_size_x)
        for y in range(grid_size_y)
        if init_grid[y][x] == 0 and goal_grid[y][x] == 0
    ]
    if not candidate_cells:
        return None
    selected = min(candidate_cells)
    return [(selected[2], selected[3])]


def main():
    args = parse()

    assert args.min_height <= args.max_height

    rng = numpy.random.default_rng(args.random_seed)

    increase_x = args.size_x < args.size_y

    plan = None
    tries = 0
    while not plan:
        tries += 1
        if (
            tries > args.ensure_plan_tries
            or args.num_towers > args.size_x * args.size_y
        ):
            tries = 0
            if increase_x:
                args.size_x += 1
            else:
                args.size_y += 1
            increase_x = not increase_x
            continue

        problem = gen_problem(args, rng)
        if problem:
            (
                grid_size_x,
                grid_size_y,
                max_height,
                init_grid,
                goal_grid,
                depots,
                robots,
            ) = problem
        else:
            continue

        if not args.ensure_plan:
            break

        plan = solve(
            grid_size_x,
            grid_size_y,
            max_height,
            robots,
            depots,
            init_grid,
            goal_grid,
            minimize=False,
        )

        # tries = 0
        # orig_size_x = grid_size_x
        # orig_size_y = grid_size_y

        # while not plan:
        #     print(tries)
        #     tries += 1
        #     # if tries == args.ensure_plan_tries:
        #     #     tries = 0

        #     if grid_size_x - orig_size_x <= grid_size_y  - orig_size_y:
        #         grid_size_x += 1
        #     else:
        #         grid_size_y += 1

    instance_name_list = ["termes"]

    # if args.ensure_plan:
    #     instance_name_list.append("{:05d}".format(len(plan)))
    # We do not consider plan length, as we are not optimizing
    problem_size = grid_size_x * grid_size_y * max_height

    instance_name_list += [
        f"{problem_size:04d}",
        "{}x{}x{}".format(
            grid_size_x, grid_size_y, max_height, args.min_height, args.num_towers
        ),
    ]

    instance_name = "-".join(instance_name_list)

    if args.output_type.startswith("pddl"):
        lines = generate_for_pddl(
            instance_name,
            grid_size_x,
            grid_size_y,
            max_height,
            robots,
            depots,
            init_grid,
            goal_grid,
            False,
        )
        if args.output_type == "pddl":
            print("\n".join(lines))
        else:
            f = open(f"instances/p-{instance_name}.pddl", "w")
            f.write("\n".join(lines))
            f.close()
    elif args.output_type == "readable":
        lines = generate_readable(
            instance_name,
            grid_size_x,
            grid_size_y,
            max_height,
            robots,
            depots,
            init_grid,
            goal_grid,
        )
        print("\n".join(lines))

    if args.store_plan:
        if not plan:
            plan = solve(
                grid_size_x,
                grid_size_y,
                max_height,
                robots,
                depots,
                init_grid,
                goal_grid,
            )
        if plan:
            f = open(f"plans/p-{instance_name}.pddl", "w")
            f.write("\n".join(plan))
            f.close()


def gen_problem(args, rng):

    grid_size_x = args.size_x
    grid_size_y = args.size_y
    max_height = args.max_height

    i_grid = [[0]]
    g_grid = gen_random_tower_boards.gen_board_autoscale(
        grid_size_x, grid_size_y, args.min_height, args.max_height, args.num_towers, rng
    )

    if not args.dont_remove_slack:
        i_grid = remove_slack(i_grid)
        g_grid = remove_slack(g_grid)

    dxinit = (grid_size_x - len(i_grid[0])) // 2 + ((grid_size_x - len(i_grid[0])) % 2)
    dyinit = (grid_size_y - len(i_grid)) // 2 + ((grid_size_y - len(i_grid)) % 2)

    dxgoal = (grid_size_x - len(g_grid[0])) // 2 + ((grid_size_x - len(g_grid[0])) % 2)
    dygoal = (grid_size_y - len(g_grid)) // 2 + ((grid_size_y - len(g_grid)) % 2)

    init_grid = [
        [get_value(i_grid, x, y, dxinit, dyinit) for x in range(0, grid_size_x)]
        for y in range(0, grid_size_y)
    ]
    goal_grid = [
        [get_value(g_grid, x, y, dxgoal, dygoal) for x in range(0, grid_size_x)]
        for y in range(0, grid_size_y)
    ]

    depots = place_depots(grid_size_x, grid_size_y, init_grid, goal_grid)
    if not depots:
        return None

    robots = [depots[0]]  # A single robot, placed in the first depot

    return grid_size_x, grid_size_y, max_height, init_grid, goal_grid, depots, robots


if __name__ == "__main__":
    main()



#! /usr/bin/env python

# Gives statistics for the instances of the Peg Solitaire suite.
# We use this to pick the 30 instances to use for the competition.
##
# Note that this script does not read the sources provided in
# instances.rb, so needs to be changed if different instances are
# used.
##
# The list of "unsolvable" instances doesn't match what is declared
# in the "diagonal moves" info for some instances, but has been
# obtained through running the planners. Apparently, the info in the
# instance descriptions has a few errors. ("Unsolvable" instances that
# aren't.)

pegs = [
    11,
    5,
    6,
    7,
    11,
    10,
    8,
    8,
    8,
    9,
    7,
    8,
    10,
    6,
    7,
    7,
    10,
    9,
    12,
    12,
    7,
    8,
    8,
    8,
    7,
    7,
    8,
    8,
    8,
    16,
    19,
    11,
    14,
    10,
    12,
    15,
    11,
    10,
    13,
    15,
    14,
    13,
    10,
    17,
    10,
    8,
    18,
    9,
    9,
    9,
    15,
    10,
    11,
    12,
    14,
    11,
    8,
    9,
    9,
    16,
    11,
    13,
    16,
    12,
    17,
    12,
    13,
    12,
    18,
    15,
    19,
    19,
    16,
    25,
    9,
    12,
    10,
    12,
    8,
    17,
    13,
    10,
    13,
    23,
    16,
    12,
    16,
    13,
    11,
    22,
    19,
    17,
    17,
    12,
    17,
    15,
    13,
    21,
    28,
    16,
    10,
    12,
    9,
    20,
    32,
]

unsolvable = [
    22,
    23,
    25,
    27,
    28,
    29,
    35,
    38,
    43,
    45,
    49,
    50,
    57,
    59,
    69,
    72,
    76,
    77,
    79,
    81,
    83,
    86,
    89,
    90,
    91,
    94,
    95,
    96,
    98,
    100,
    103,
]


def report(instances):
    instances_and_pegs = [(instance, pegs[instance - 1]) for instance in instances]
    instances_and_pegs.sort(key=lambda x: (x[1], x[0]))
    for pair in instances_and_pegs:
        print("instance #%03d: %2d pegs" % pair)


solvable = set(range(1, len(pegs) + 1)) - set(unsolvable)
print("Solvable instances (%d):" % len(solvable))
report(solvable)
print()
print("Unsolvable instances (%d):" % len(unsolvable))
report(unsolvable)



#!/usr/bin/env python

import argparse
from itertools import product
import random

import sys


def generate_for_solver(instance_name, num_decks, num_suits, num_values, s):
    lines = []
    deals, piles = s
    lines.append(f"{num_decks} {num_suits} {num_values}")

    lines.append(str(len(deals)))
    for deal in deals:
        lines.append(" ".join([str(len(deal))] + ["%s %s %s" % card for card in deal]))

    lines.append(str(len(piles)))
    for card_stack in piles:
        lines.append(
            " ".join(
                [str(len(card_stack))] + ["%s %s %s" % card for card in card_stack]
            )
        )

    return lines


def generate_readable(instance_name, num_decks, num_suits, num_values, s):
    lines = []
    deals, piles = s
    lines += [
        "Using %s decks of cards with %s suits per deck and %s values per suit"
        % (num_decks, num_suits, num_values)
    ]
    lines += ["", "Deals"]
    deck = [c for deal in deals for c in deal]
    for i, deal in enumerate(deals):
        lines += [
            "deal %s: " % i
            + " ".join(
                [f"d{deck}-s{suite}-v{value}" for deck, suite, value in deal]
            )
        ]
    lines += ["", "Initial configuration of piles"]
    for i, pile in enumerate(piles):
        lines += [
            "pile %s: " % i
            + " ".join(
                [f"d{deck}-s{suite}-v{value}" for deck, suite, value in pile]
            )
        ]
    return lines


def generate_for_pddl(instance_name, num_decks, num_suits, num_values, s):
    deals, piles = s

    comments = generate_readable(instance_name, num_decks, num_suits, num_values, s)

    objects = []
    objects += [
        f"card-d{deck}-s{suit}-v{value} - card"
        for deck in range(num_decks)
        for suit in range(num_suits)
        for value in range(num_values)
    ]
    objects += ["pile-%s - tableau" % pile_id for pile_id in range(len(piles))]
    objects += ["deal-%s - deal" % deal_id for deal_id in range(len(deals) + 1)]

    facts = []
    # Cards in tableau piles
    for pile_id, pile in enumerate(piles):
        facts += [
            "(on card-d%s-s%s-v%s card-d%s-s%s-v%s)" % (card1 + card2)
            for card1, card2 in zip(pile[1:], pile)
        ]
        facts += ["(on card-d%s-s%s-v%s pile-%s)" % (pile[0] + (pile_id,))]
        facts += ["(clear card-d%s-s%s-v%s)" % pile[-1]]
        facts += [f"(part-of-tableau pile-{pile_id} pile-{pile_id})"]
        facts += [
            "(part-of-tableau card-d%s-s%s-v%s pile-%s)" % (card + (pile_id,))
            for card in pile
        ]
        pile_parts = split_pile_in_movable_parts(pile)
        facts += ["(movable card-d%s-s%s-v%s)" % card for card in pile_parts[-1]]
        facts += ["(in-play card-d%s-s%s-v%s)" % card for card in pile]

    # Cards in deals
    for deal_id, deal in enumerate(deals):
        deal_deck = list(reversed(deal))
        facts += ["(on card-d%s-s%s-v%s deal-%s)" % (deal_deck[-1] + (deal_id,))]
        facts += [
            "(on card-d%s-s%s-v%s card-d%s-s%s-v%s)" % (card2 + card1)
            for card1, card2 in zip(deal_deck[1:], deal_deck)
        ][::-1]
        facts += ["(clear card-d%s-s%s-v%s)" % deal_deck[0]]

    facts += ["(current-deal deal-0)"]

    # Static predicates
    facts += [
        "(CAN-CONTINUE-GROUP card-d%s-s%s-v%s card-d%s-s%s-v%s)"
        % (deck1, suit, value, deck2, suit, value + 1)
        for deck1 in range(num_decks)
        for deck2 in range(num_decks)
        for suit in range(num_suits)
        for value in range(num_values - 1)
    ]
    facts += [
        "(CAN-BE-PLACED-ON card-d%s-s%s-v%s card-d%s-s%s-v%s)"
        % (deck1, suit1, value, deck2, suit2, value + 1)
        for deck1 in range(num_decks)
        for deck2 in range(num_decks)
        for suit1 in range(num_suits)
        for suit2 in range(num_suits)
        for value in range(num_values - 1)
    ]
    facts += [
        f"(IS-ACE card-d{deck}-s{suit}-v0)"
        for deck in range(num_decks)
        for suit in range(num_suits)
    ]
    facts += [
        "(IS-KING card-d%s-s%s-v%d)" % (deck, suit, num_values - 1)
        for deck in range(num_decks)
        for suit in range(num_suits)
    ]
    facts += [
        f"(NEXT-DEAL deal-{deal_id} deal-{deal_id + 1})"
        for deal_id in range(len(deals))
    ]
    for deal_id, deal in enumerate(deals):
        deal_deck = list(reversed(deal))
        facts += [
            "(TO-DEAL card-d%s-s%s-v%s pile-%s deal-%s deal-%s)"
            % (deal_deck[-1] + (len(deal) - 1, deal_id, deal_id))
        ]
        facts += [
            "(TO-DEAL card-d%s-s%s-v%s pile-%s deal-%s card-d%s-s%s-v%s)"
            % (nextcard + (pile_id, deal_id) + card)
            for pile_id, (card, nextcard) in enumerate(zip(deal_deck[1:], deal_deck))
        ]

    facts += ["(= (total-cost) 0)"]

    goals = (
        ["(clear pile-%s)" % pile_id for pile_id in range(len(piles))]
        + ["(clear deal-%s)" % deal_id for deal_id in range(len(deals))]
        + [
            f"(on card-d{deck}-s{suit}-v{value} discard)"
            for deck in range(num_decks)
            for suit in range(num_suits)
            for value in range(num_values)
        ]
    )

    lines = []
    lines += ["(define (problem %s)" % instance_name]
    lines += ["(:domain spider)"]
    lines += ["; %s" % c for c in comments]
    lines += ["(:objects"]
    lines += ["    %s" % o for o in objects]
    lines += [")", "(:init"]
    lines += ["    %s" % f for f in facts]
    lines += [")", "(:goal (and"]
    lines += ["    %s" % g for g in goals]
    lines += ["))", "(:metric minimize (total-cost))"]
    lines += [")"]
    return lines


def is_compatible(card1, card2):
    _, suit1, value1 = card1
    _, suit2, value2 = card2
    return suit1 == suit2 and value1 == value2 + 1


def split_pile_in_movable_parts(pile):
    parts = []
    current_part = []
    for card in pile:
        if current_part and not is_compatible(current_part[-1], card):
            parts.append(current_part)
            current_part = []
        current_part.append(card)
    parts.append(current_part)
    return parts


def random_state(num_decks, num_suits, num_values, num_piles, num_deals):
    cards = list(product(range(num_decks), range(num_suits), range(num_values)))
    random.shuffle(cards)
    # use the first cards for the deals
    deals = [cards[i * num_piles : (i + 1) * num_piles] for i in range(num_deals)]
    # distribute remaining cards as evenly as possible
    cards = cards[num_deals * num_piles :]
    num_cards_per_pile = int(len(cards) / num_piles)

    piles = [
        cards[i * num_cards_per_pile : (i + 1) * num_cards_per_pile]
        for i in range(num_piles)
    ]
    # distribute remaining cards to the first piles
    cards = cards[num_piles * num_cards_per_pile :]
    for pile, card in zip(piles, cards):
        pile.append(card)
    return deals, piles


def parse():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "num_decks", type=int, help="Number of decks of cards used for the game."
    )
    parser.add_argument(
        "num_suits", type=int, help="Number of different suits of cards in each deck."
    )
    parser.add_argument(
        "num_values",
        type=int,
        help="Number of different values of each suit in each deck.",
    )
    parser.add_argument("num_piles", type=int, help="Number of piles.")
    parser.add_argument(
        "num_deals",
        type=int,
        help="Number of deals (each deal contains one card for each pile, remaining cards are distributed to the piles).",
    )
    parser.add_argument("random_seed", type=int)
    parser.add_argument("output_type", choices=["pddl", "raw", "readable"])
    return parser.parse_args()


def main():
    args = parse()
    instance_name = "spider-{}-{}-{}-{}-{}-{}".format(
        args.num_decks,
        args.num_suits,
        args.num_values,
        args.num_piles,
        args.num_deals,
        args.random_seed,
    )

    random.seed(args.random_seed)
    s = random_state(
        args.num_decks, args.num_suits, args.num_values, args.num_piles, args.num_deals
    )
    if args.output_type == "raw":
        lines = generate_for_solver(
            instance_name, args.num_decks, args.num_suits, args.num_values, s
        )
    elif args.output_type == "pddl":
        lines = generate_for_pddl(
            instance_name, args.num_decks, args.num_suits, args.num_values, s
        )
    elif args.output_type == "readable":
        lines = generate_readable(
            instance_name, args.num_decks, args.num_suits, args.num_values, s
        )
    print("\n".join(lines))


if __name__ == "__main__":
    main()




import itertools
import math
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from pddl_plus_parser.lisp_parsers import ProblemParser, TrajectoryParser
from torch.utils.data import DataLoader, TensorDataset
import numpy as np
from experiment_runner.rosame_runner import Rosame_Runner
import sys
from pathlib import Path
from pddl_plus_parser.models import Domain, Observation, State, ActionCall, PDDLFunction, Predicate, GroundedPredicate, create_type_hierarchy_graph

from pddl_plus_parser.lisp_parsers import ProblemParser, TrajectoryParser, DomainParser

sys.path.append("/Users/omarwattad/Documents/Action-Model-Research/sam_learning")
from sam_learning.learners.multi_agent_sam import MultiAgentSAM

def prepare_rosame_data(model, observation):
    """
    Prepares structured data from traces to be used within the ROSAME framework, which the data should be encoded to True/False (0  and 1)
    using the triples that are created using macq every fluent is true takes 1 else 0
    returns:
    - encoded pre_state
    - encoded next_state
    - encoded action
    """
    steps_action = []
    steps_state1, steps_state2 = [], []
    for component in observation.components:
        exec_action = []
        for action in component.grounded_joint_action.actions:
            action_lst = action.__str__()[1:-1].split()

            if len(action_lst) != len(set(action_lst)):
                continue
            if action.__str__()[1:-1] == 'nop ':
                exec_action.append(-1)
            else:
                exec_action.append(model.rosame.actions[action.__str__()[1:-1]])
        steps_action.append(exec_action)
        # if not steps_action: continue
        next_state, pre_state = [], []
        for _, val in component.next_state.state_predicates.items():
            for pred in val:
                next_state.append(pred.untyped_representation[1:-1])
        for _, val in component.previous_state.state_predicates.items():
            for pred in val:
                pre_state.append(pred.untyped_representation[1:-1])
        state1 = [1 if p in pre_state else 0 for p in model.rosame.propositions]
        state2 = [1 if p in next_state else 0 for p in model.rosame.propositions]
        steps_state1.append(state1)
        steps_state2.append(state2)
    return steps_state1,steps_action, steps_state2

def learn_rosame(model, observation):
    """Learns an action model using ROSAME learning with multi-agent joint actions (no 'nop')."""

    # Each step has: state1, joint_action (list of ints), state2
    steps_state1, steps_joint_action, steps_state2 = prepare_rosame_data(model,observation)

    steps_state1_tensor = torch.tensor(np.array(steps_state1)).float()           # [N, num_props]
    steps_joint_action_tensor = torch.tensor(np.array(steps_joint_action))       # [N, num_agents]
    steps_state2_tensor = torch.tensor(np.array(steps_state2)).float()           # [N, num_props]

    batch_sz = 1000
    dataset = TensorDataset(steps_state1_tensor, steps_joint_action_tensor, steps_state2_tensor)
    dataloader = DataLoader(dataset, batch_size=batch_sz, shuffle=False)

    # Collect learnable parameters from all schemas
    parameters = [{'params': schema.parameters(), 'lr': 1e-3} for schema in model.rosame.action_schemas]
    optimizer = optim.Adam(parameters)

    for epoch in range(100):
        epoch_loss = 0.0

        for i, (state_1, joint_actions, state_2) in enumerate(dataloader):
            optimizer.zero_grad()
            total_loss = 0.0

            for agent_idx in range(joint_actions.shape[1]):
                # Get actions for this agent in the batch
                actions = joint_actions[:, agent_idx]
                # actions = actions[actions != -1]
                # print(actions)# [batch_size]
                # Build precondition/add/delete matrices for those actions
                precon, addeff, deleff = model.rosame.build(actions)

                # Predict next state based on effects applied to current state
                predicted_state2 = state_1 * (1 - deleff) + (1 - state_1) * addeff

                # Loss between predicted and actual next state
                effect_loss = F.mse_loss(predicted_state2, state_2, reduction='sum')

                # Precondition validity loss (should not violate preconditions)
                validity_constraint = (1 - state_1) * precon
                precon_loss = F.mse_loss(validity_constraint,
                                         torch.zeros_like(validity_constraint),
                                         reduction='sum')

                # Encourage model to learn 1s in preconditions
                reg_loss = 0.2 * F.mse_loss(precon, torch.ones_like(precon), reduction='sum')

                total_loss += effect_loss + precon_loss + reg_loss

            total_loss.backward()
            optimizer.step()

            epoch_loss += total_loss.item() / batch_sz

        if epoch % 10 == 0:
            print(f"[Epoch {epoch}] Avg Loss: {epoch_loss:.6f}")

dataset_path = Path("/Users/omarwattad/Downloads/ma-sam-main/experiments_dataset")
algo = Rosame_Runner(dataset_path / "satellite/satellite_combined_domain.pddl")
partial_domain = DomainParser(dataset_path / "satellite/satellite_combined_domain.pddl",partial_parsing=True).parse_domain()
print(partial_domain.to_pddl())
observations = []
counter = 0
for traj in [5,6,7,8,9,10,11,12,13,14,15,16,18,19,20]:
    if traj > 9:
        problem = ProblemParser(dataset_path / f"satellite/p{traj}-pfile{traj}.pddl", partial_domain).parse_problem()
    else:
        problem = ProblemParser(dataset_path / f"satellite/p0{traj}-pfile{traj}.pddl", partial_domain).parse_problem()

    if traj > 9:
        observation = TrajectoryParser(partial_domain, problem).parse_trajectory(
            trajectory_file_path=dataset_path / f"satellite/p{traj}-pfile{traj}.trajectory",
            executing_agents=["satellite0", "satellite1", "satellite2", "satellite3", "satellite4", "satellite5",
                              "satellite6", "satellite7", "satellite8", "satellite9"])
    else:
        observation = TrajectoryParser(partial_domain,problem).parse_trajectory(trajectory_file_path=dataset_path / f"satellite/p0{traj}-pfile{traj}.trajectory",executing_agents=["satellite0","satellite1","satellite2","satellite3","satellite4","satellite5","satellite6","satellite7","satellite8","satellite9"])

    observations.append(observation)
    algo.add_problem(problem)
    algo.ground_new_trajectory()
    learn_rosame(algo, observation)
    counter += 1
    sam = MultiAgentSAM(partial_domain)
    learned, _ = sam.learn_combined_action_model(observations)
    with open(f"/Users/omarwattad/Documents/Action-Model-Research/rosame/MA_domains/satellite/MA-SAM_{counter}", "w") as f:
        f.write(learned.to_pddl())
    algo.export_rosame_domain(f"/Users/omarwattad/Documents/Action-Model-Research/rosame/MA_domains/satellite/MA-ROSAME_{counter}.pddl")


import numpy

FILE_SIZE_DISTRIBUTION = lambda: numpy.random.random_integers(1, 5)

SERVERS = {
    "server1": {
        "capacity": 16,
        "io": 5,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(20, 4)))),
    },
    "server2": {
        "capacity": 8,
        "io": 1,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(10, 2)))),
    },
    "server3": {
        "capacity": 8,
        "io": 1,
        "processing_distr": (lambda: max(1, int(numpy.random.normal(10, 2)))),
    },
}

CONNECTIONS = {(1, 2): {"send": 4}, (1, 3): {"send": 6}}



import abc
from dataclasses import dataclass
from typing import List


@dataclass
class OfflineAlgorithmAdapter(abc.ABC):
    """
    An abstract class for an action model learning algorithm, which defines the abstract interface that must be
    implemented by every (subclass) algorithm adapter to enable automated evaluation.
    """

    @abc.abstractmethod
    def learn(self,
              domain_path: str,
              trajectory_paths: List[str]) -> str:
        """
        Learns a PDDL action model from:
         (i)    a (possibly empty) input model which is required to specify the predicates and operators signature;
         (ii)   a list of trajectory file paths.

        :parameter domain_path: input PDDL domain file path
        :parameter trajectory_paths: list of trajectory file paths

        :return: a string representing the learned PDDL model
        """
        raise NotImplementedError



#! /usr/bin/env python

# CITYCAR DOMAIN GENERATOR.
# IPC 2014
# author: Mauro Vallati -- University of Huddersfield


import sys
import random
import argparse


to_print_car_pos = ""

parser = argparse.ArgumentParser()

parser.add_argument("rows",type=int,help="the number of grid rows.")
parser.add_argument("columns",type=int,help="the number of grid columns.")
parser.add_argument("cars",type=int,help="how many cars have to go through the network")
parser.add_argument("garages",type=int,help="the number of starting garages")
parser.add_argument("--density",type=float,default=1.0,help="The ratio of the available roads in the road network (0.0 <= density <= 1.0). Junctions are randomly set obstructed/unavailable for passage")
parser.add_argument("--seed",type=int,help="random seed")

args = parser.parse_args()

row     = args.rows
column  = args.columns
car     = args.cars
garage  = args.garages
density = args.density
seed    = args.seed

if not (0.0 <= density <= 1.0):
    parser.error("density must be a probability (0.0 <= density <= 1.0)")


random.seed(seed)

road = int(sys.argv[1]) + 2

matrix = [["free" for i in range(column)] for j in range(row)]


for posizionate in range(car):
    x = random.randint(0, garage - 1)
    to_print_car_pos = (
        to_print_car_pos
        + "(starting car"
        + str(posizionate)
        + " garage"
        + str(x)
        + ")\n"
    )


# Finally, time for printing everything out.

print("(define (problem citycar-" + str(row) + "-" + str(column) + "-" + str(car) + ")")
print("(:domain citycar)")
print("(:objects ", end=" ")

for i in range(row):
    print("")
    for j in range(column):
        x = "junction" + str(i) + "-" + str(j)
        print(x, end=" ")
print("- junction")

for i in range(car):
    x = "car" + str(i)
    print(x, end=" ")
print("- car")

for i in range(garage):
    x = "garage" + str(i)
    print(x, end=" ")
print("- garage")

for i in range(road):
    x = "road" + str(i)
    print(x, end=" ")
print("- road")

print(")\n(:init")

# connected cells on the same line
for i in range(row):
    for j in range(column - 1):
        print(
            "(same_line junction"
            + str(i)
            + "-"
            + str(j)
            + " junction"
            + str(i)
            + "-"
            + str(j + 1)
            + ")"
        )
        print(
            "(same_line junction"
            + str(i)
            + "-"
            + str(j + 1)
            + " junction"
            + str(i)
            + "-"
            + str(j)
            + ")"
        )

# connected cells on the same column
for i in range(column):
    for j in range(row - 1):
        print(
            "(same_line junction"
            + str(j)
            + "-"
            + str(i)
            + " junction"
            + str(j + 1)
            + "-"
            + str(i)
            + ")"
        )
        print(
            "(same_line junction"
            + str(j + 1)
            + "-"
            + str(i)
            + " junction"
            + str(j)
            + "-"
            + str(i)
            + ")"
        )

# make the diagonal connections
for i in range(row - 1):
    for j in range(column - 1):
        print(
            "(diagonal junction"
            + str(i)
            + "-"
            + str(j)
            + " junction"
            + str(i + 1)
            + "-"
            + str(j + 1)
            + ")"
        )
        print(
            "(diagonal junction"
            + str(i + 1)
            + "-"
            + str(j + 1)
            + " junction"
            + str(i)
            + "-"
            + str(j)
            + ")"
        )
        print(
            "(diagonal junction"
            + str(i)
            + "-"
            + str(j + 1)
            + " junction"
            + str(i + 1)
            + "-"
            + str(j)
            + ")"
        )
        print(
            "(diagonal junction"
            + str(i + 1)
            + "-"
            + str(j)
            + " junction"
            + str(i)
            + "-"
            + str(j + 1)
            + ")"
        )

# free cells
for i in range(row):
    for j in range(column):
        if density != 1.0 and i != 0 and i < row - 1 and j != 0 and j < column - 1:
            if density > random.random():
                print("(clear junction" + str(i) + "-" + str(j) + ")")
        else:
            print("(clear junction" + str(i) + "-" + str(j) + ")")

for i in range(garage):
    x = random.randint(0, (column - 1))
    print("(at_garage garage" + str(i) + " junction0" + "-" + str(x) + ")")

# occupied cells
print(to_print_car_pos, end=" ")
print("(= (total-cost) 0)")
print(")\n(:goal\n(and")

for i in range(car):
    x = random.randint(0, (column - 1))
    print("(arrived car" + str(i) + " junction" + str(row - 1) + "-" + str(x) + ")")

print(")\n)\n(:metric minimize (total-cost))\n)")



import pandas as pd
from pddl_plus_parser.lisp_parsers import DomainParser, ProblemParser, TrajectoryParser
from typing import List, Union, Optional, Dict

class ExperimentRunner():
    def __init__(self, domain_file= None):
        self.gold_standard = DomainParser(domain_file, partial_parsing=False).parse_domain()
        self.statistics = {"Algorithm": [],
                           "Domain": [],
                           "Num. Of Traces": [],
                           "Threshold": [],
                           "Epsilon": [],
                           "Fold": [],
                           'Epoch': [],
                           'Unobserved_actions': [],
                           "precision_effect":[],
                           'precision_precondition':[],
                           'recall_effect':[],
                           'recall_precondition':[]}

    def export_statistics(self, path):
        df = pd.DataFrame(self.statistics)
        df.to_csv(path,index=False)


    def creat_object_map(self, params, learned_params):
        """
        Creates a mapping between object parameters and their assigned indices.

        This function generates a dictionary that maps each parameter from two given lists
        (`params` and `learned_params`) to a unique string representation of its index.
        The index is assigned based on the parameter's position in the input lists.

        Args:
            params (List[str]): A list of parameter names.
            learned_params (List[str]): A list of corresponding learned parameter names.

        Returns:
            Dict[str, str]: A dictionary mapping each parameter (from both lists) to a string index.

        """
        params_map = {}
        for i in range(len(params)):
            params_map[params[i]] = str(i)
            params_map[learned_params[i]] = str(i)
        return params_map

    def get_predicates(self,predicates,params_map): # TODO: think about a better way to compare predicates other that mapping the predicates to indices.
        preds = []
        for predicate in predicates:
            predicate_str = ''
            if not predicate.is_positive:
                predicate_str += "not "
            predicate_str += predicate.name + " "
            param_lst = []
            for param in predicate.signature:
                param_lst.append(params_map[param])
            predicate_str += " ".join(param_lst)
            preds.append(predicate_str)
        return preds


    def run_single_experiment(self, learned_domain_path, algorithm_name, num_traces, threshold="NA", eps="NA",meta_data = None,fold=0,epoch="NA"):
        """
         Compares the preconditions and effects of actions between a learned PDDL domain and a gold standard domain.

        This function compares the actions, preconditions, and effects of actions from a learned domain
        with a gold standard domain. It calculates the precision, recall, and false positive/false negative
        values for preconditions and effects. It also prints warnings for mismatches in preconditions and effects
        and stores statistical values in `self.statistics`.

        Args:
            learned_domain_path (str): The path to the PDDL file containing the learned domain to compare.
            algorithm_name (str): The name of the algorithm used to learn the domain.
            num_traces (int): The number of traces used in learning the domain.

        Returns:
            None: The function updates the `self.statistics` dictionary with precision, recall, and other metrics.
        """
        learned_domain = DomainParser(learned_domain_path, partial_parsing=False).parse_domain()
        precision_pre, recall_pre, precision_eff, recall_eff = 0 ,0 ,0 ,0
        unobserved = 0

        """Compare preconditions and effects between actions of two PDDL domains."""
        for action in self.gold_standard.actions:
            FP_pre, FN_pre, FP_eff, FN_eff = 0, 0, 0, 0
            TP_pre, TP_eff = 0, 0  # True positives for preconditions and effects


            if action not in learned_domain.actions:
                continue
            if meta_data and meta_data[action] == 'UNOBSERVED':
                precision_eff += 1
                precision_pre += 0
                recall_eff += 0
                recall_pre += 1
                unobserved += 1
                print(f'Action {action} is unobserved.')
                continue
            params_map = self.creat_object_map(self.gold_standard.actions[action].parameter_names,learned_domain.actions[action].parameter_names)
            gold_standard_precond = self.gold_standard.actions[action].preconditions.root.operands
            learned_precond = learned_domain.actions[action].preconditions.root.operands
            new_precond = []
            for predicate in gold_standard_precond:
                if predicate.is_positive:
                    new_precond.append(predicate)
            # gold_standard_precond = self.get_predicates(new_precond, params_map)
            gold_standard_precond = new_precond

            new_precond = []
            for predicate in learned_precond:
                if predicate.is_positive:
                    new_precond.append(predicate)
            # learned_precond = self.get_predicates(new_precond, params_map)
            learned_precond = new_precond
            # Compare precondition order
            if gold_standard_precond != learned_precond:
                print(f"Warning: The preconditions of action {action} have a different order.")

            # True Positives (matching preconditions)
            precond1_set = set(gold_standard_precond)
            TP_pre += len(precond1_set.intersection(set(learned_precond)))

            # False negatives (missing preconditions)
            missing_preconditions = set(gold_standard_precond) - set(learned_precond)
            if missing_preconditions:
                print(f"FALSE NEGATIVE Warning: Missing preconditions in second domain for action {action}: {[p.untyped_representation for p in missing_preconditions]}")
                FN_pre += len(missing_preconditions)

            # False positives (extra preconditions)
            extra_preconditions = set(learned_precond) - set(gold_standard_precond)
            if extra_preconditions:
                print(f"FALSE POSITIVE Warning: Extra preconditions in second domain for action {action}: {[p.untyped_representation for p in extra_preconditions]}")
                FP_pre += len(extra_preconditions)

            # Compare effects similarly
            # gold_effects = self.get_predicates(self.gold_standard.actions[action].discrete_effects, params_map)
            # learned_effects = self.get_predicates(learned_domain.actions[action].discrete_effects, params_map)
            gold_effects = self.gold_standard.actions[action].discrete_effects
            learned_effects = learned_domain.actions[action].discrete_effects
            # Compare effect order
            if gold_effects != learned_effects:
                print(f"Warning: The effects of action {action} have a different order.")

            # True Positives (matching effects)
            effects1_set = set(gold_effects)
            TP_eff += len(effects1_set.intersection(set(learned_effects)))

            # False negatives (missing effects)
            missing_effects = set(gold_effects) - set(learned_effects)
            if missing_effects:
                print(f"FALSE NEGATIVE Warning: Missing effects in second domain for action {action}: {[p.untyped_representation for p in missing_effects]}")
                FN_eff += len(missing_effects)

            # False positives (extra effects)
            extra_effects = set(learned_effects) - set(gold_effects)
            if extra_effects:
                print(f"FALSE POSITIVE Warning: Extra effects in second domain for action {action}: {[p.untyped_representation for p in extra_effects]}")
                FP_eff += len(extra_effects)


            precision_pre += TP_pre / (TP_pre + FP_pre) if (TP_pre + FP_pre) > 0 else 0
            recall_pre += TP_pre / (TP_pre + FN_pre) if (TP_pre + FN_pre) > 0 else 0

            # Precision and Recall for Effects
            precision_eff += TP_eff / (TP_eff + FP_eff) if (TP_eff + FP_eff) > 0 else 0
            recall_eff += TP_eff / (TP_eff + FN_eff) if (TP_eff + FN_eff) > 0 else 0
        num_of_actions = len(self.gold_standard.actions)
        self.statistics["precision_effect"].append(round(precision_eff / num_of_actions, 2))
        self.statistics["precision_precondition"].append(round(precision_pre / num_of_actions, 2))
        self.statistics["recall_effect"].append(round(recall_eff / num_of_actions, 2))
        self.statistics["recall_precondition"].append(round(recall_pre / num_of_actions, 2))
        self.statistics["Threshold"].append(threshold)
        self.statistics["Algorithm"].append(algorithm_name)
        self.statistics['Domain'].append(self.gold_standard.name)
        self.statistics["Num. Of Traces"].append(num_traces)
        self.statistics["Epsilon"].append(eps)
        self.statistics['Fold'].append(fold)
        self.statistics['Unobserved_actions'].append(unobserved)
        self.statistics['Epoch'].append(epoch)




        # Print False Positive, False Negative, Precision, and Recall results ## TODO: REMOVE later!
        print("---------------------------------------------------------------------------------------")
        print(f"False Positives: Preconditions: {FP_pre}, Effects: {FP_eff}")
        print(f"False Negatives: Preconditions: {FN_pre}, Effects: {FN_eff}")
        print(f"Precision: Preconditions: {precision_pre / num_of_actions:.2f}, Effects: {precision_eff / num_of_actions:.2f}")
        print(f"Recall: Preconditions: {recall_pre / num_of_actions:.2f}, Effects: {recall_eff / num_of_actions:.2f}")
        print("---------------------------------------------------------------------------------------")

    def run_experiment(self,algo_lst:List[str],domain_path_lst:List[str], trace_len_lst:List[int], threshold_lst:List[float]= [], eps_lst:List[float] = []):
        for algo in algo_lst:
            for domain_path in domain_path_lst:
                for trace_len in trace_len_lst:
                    # for threshold in threshold_lst:
                    #     for eps in eps_lst:
                    self.run_single_experiment(algo, domain_path, trace_len)


#!/usr/bin/env python

import argparse
from itertools import product
import random

import sys


# Board notation:
#
class Board:
    def __init__(self, board_name):
        if board_name.startswith("empty"):
            width, height = (
                board_name.replace("empty-", "").replace("empty", "").split("x")
            )
            self.board = [
                ["_" for _ in range(0, int(height))] for _ in range(0, int(width))
            ]
        else:
            f = open(board_name)
            self.board = [[a for a in l if a != "\n"] for l in f.readlines()]
            f.close()

    def __repr__(self):
        return "\n".join(map(lambda x: "".join(x), self.board))

    def width(self):
        return len(self.board)

    def height(self):
        return len(self.board[0])

    def num_clear_positions(self):
        positions = [
            (i, j)
            for (i, j) in product(range(self.width()), range(self.height()))
            if self.board[i][j] == "_"
        ]

        return len(positions)

    def random_clear_positions(self):
        positions = [
            (i, j)
            for (i, j) in product(range(self.width()), range(self.height()))
            if self.board[i][j] == "_"
        ]
        random.shuffle(positions)
        return positions

    def blocked_positions(self):
        positions = [
            (i, j)
            for (i, j) in product(range(self.width()), range(self.height()))
            if self.board[i][j] == "S"
            or self.board[i][j] == "H"
            or self.board[i][j] == "*"
        ]
        return positions

    def generate_apples(self, num_apples, mark="A"):
        # apples = [(i, j) for (i, j) in product(range(self.width()), range(self.height())) if self.board[i][j] == "A"]
        new_apples = self.random_clear_positions()[0:num_apples]
        for (x, y) in new_apples:
            self.board[x][y] = mark
        return new_apples

    def is_empty(self, x, y):
        if x < 0 or x >= self.width():
            return False
        if y < 0 or y >= self.height():
            return False

        return self.board[x][y] == "_"

    def generate_snake_aux(self, x, y, snake_size, snake):

        if not self.is_empty(x, y):
            return False

        self.board[x][y] = "S"

        if snake_size == 0:
            snake += [(x, y)]
            return True

        pos_list = [(x + 1, y), (x - 1, y), (x, y + 1), (x, y - 1)]
        random.shuffle(pos_list)
        for pos in pos_list:
            if self.generate_snake_aux(pos[0], pos[1], snake_size - 1, snake):
                snake += [(x, y)]
                return True
        self.board[auxx][auxy] = "_"
        return False

    def generate_snake(self, snake_size):
        snake = []
        for (x, y) in self.random_clear_positions():
            if self.generate_snake_aux(x, y, snake_size, snake):
                self.board[x][y] = "H"
                return snake[::-1]

            # next_pos = []
            # if auxx > 0 and auxx <= x  and auxy == y:
            #     auxx -= 1
            # elif auxy >= y and auxy <= self.height() - 1  and auxx == self.width() - 1:
            #     auxy += 1

            # elif auxy <= y:
            #     if auxx == 0 and auxy > 0:
            #         auxy -= 1
            #     elif auxy % 2 == 0 and auxx <= self.width() - 2:
            #         auxx += 1
            #     elif auxy % 2 == 1 and auxx > 1:
            #         auxx -= 1
            #     else:
            #         auxy += 1
            # else:
            #     if auxy % 2 == self.height() % 2 and auxx > 0:
            #         auxx -= 1
            #     elif auxy % 2 != self.height() % 2 and auxx < self.width()  -2:
            #         auxx += 1
            #     else:
            #         auxy -= 1
            # assert (not (auxx, auxy) in snake)
            # snake.append((auxx, auxy))
            # self.board[auxx][auxy] = 'S'
        return snake


def print_list(l):
    print(" ".join([str(len(l))] + [f"{a} {b}" for (a, b) in l]))


def print_solver(instance_name, board, snake, apples, spawn_apples):
    print(f"{board.width()} {board.height()}")
    # print_list (board.blocked_positions())
    print_list(snake)
    print_list(apples)
    print_list(spawn_apples)


def print_readable(instance_name, board, snake, apples, spawn_apples):
    print(instance_name)
    print("Blocked: ", board.blocked_positions())
    print("Snake: ", snake)
    print("Apples: ", apples)
    print("Spawn: ", spawn_apples)


def get_adjacent_pairs(width, height):
    res = []
    for x in range(0, width):
        for y in range(0, height):
            if x < width - 1:
                res.append(((x, y), (x + 1, y)))
            if y < height - 1:
                res.append(((x, y), (x, y + 1)))
            if x > 0:
                res.append(((x, y), (x - 1, y)))
            if y > 0:
                res.append(((x, y), (x, y - 1)))
    return res


def pos_name(p):
    return "pos{}-{}".format(*p)


def print_pddl(instance_name, board, snake, apples, spawn_apples):
    positions = list(product(range(board.width()), range(board.height())))
    pos_objects = " ".join([pos_name(pos) for pos in positions])
    is_adjacent_facts = "\n ".join(
        [
            f"(isAdjacent {pos_name(adj1)} {pos_name(adj2)})"
            for (adj1, adj2) in get_adjacent_pairs(board.width(), board.height())
        ]
    )

    head_snake = f"(headSnake {pos_name(snake[0])})"
    tail_snake = f"(tailSnake {pos_name(snake[-1])})"

    next_snake_facts = "\n ".join(
        [
            f"(nextSnake {pos_name(snake[i])} {pos_name(snake[i + 1])})"
            for i in range(0, len(snake) - 1)
        ]
    )
    blocked_facts = "\n ".join(
        [f"(blocked {pos_name(sn)})" for sn in board.blocked_positions()]
    )

    is_point_facts = "\n ".join([f"(isPoint {pos_name(a)})" for a in apples])

    spawn_apples_facts = []
    if spawn_apples:
        spawn_apples_facts = [f"(spawn {pos_name(spawn_apples[0])})"] + [
            f"(nextSpawn {pos_name(spawn_apples[-1])} dummyPoint)"
        ]
        if len(spawn_apples) > 1:
            spawn_apples_facts += [
                "(nextSpawn {} {})".format(
                    pos_name(spawn_apples[i]), pos_name(spawn_apples[i + 1])
                )
                for i in range(0, len(spawn_apples) - 1)
            ]
    else:
        spawn_apples_facts = ["(spawn dummyPoint)"]

    next_point_facts = "\n ".join(spawn_apples_facts)

    goal_facts = "\n ".join(
        [f"(not (isPoint {pos_name(a)}))" for a in apples + spawn_apples]
    )

    print(
        """(define (problem {instance_name})
(:domain snake)
(:objects
    {pos_objects}
)
(:init
    {is_adjacent_facts}
    {tail_snake}
    {head_snake}
    {next_snake_facts}
    {blocked_facts}
    {next_point_facts}
    {is_point_facts}
)
(:goal (and
    {goal_facts}
))
)
""".format(
            **locals()
        )
    )


def parse():
    parser = argparse.ArgumentParser()
    parser.add_argument("board", help="What kind of board.")
    parser.add_argument("snake_size", type=int, help="Initial size of the snake.")
    parser.add_argument(
        "num_initial_apples",
        type=int,
        help="Number of apples that are present at the beginning.",
    )
    parser.add_argument(
        "num_spawn_apples", help="Number of apples that will spawn at random locations."
    )
    parser.add_argument("random_seed", type=int)
    parser.add_argument("output_type", choices=["pddl", "raw", "readable", "draw"])

    parser.add_argument("--output", default=None)

    return parser.parse_args()


def main():
    args = parse()
    board_name = args.board.replace(".txt", "")

    random.seed(args.random_seed)
    board = Board(args.board)

    if "/" in board_name:
        board_name = board_name.split("/")[-1]

    num_initial_apples = args.num_initial_apples
    if "%" in args.num_spawn_apples:
        percentage = int(args.num_spawn_apples[:-1]) / 100.0
        num_spawn_apples = (
            int(board.num_clear_positions() * percentage)
            - 1
            - args.snake_size
            - args.num_initial_apples
        )
        if num_spawn_apples < 1:  # At least one apple should be spawn
            num_initial_apples += num_spawn_apples - 1
            num_spawn_apples = 1

        # There must be at least one apple
        if num_initial_apples < 1:
            num_initial_apples = 1
    else:
        num_spawn_apples = int(args.num_spawn_apples)

    if (
        board.num_clear_positions() % 2 == 1
        and num_spawn_apples
        == int(board.num_clear_positions() * percentage)
        - 1
        - args.snake_size
        - args.num_initial_apples
    ):
        num_spawn_apples -= 1

    instance_name = "snake-%s-%s-%s-%d-%s" % (
        board_name,
        args.snake_size,
        num_initial_apples,
        num_spawn_apples,
        args.random_seed,
    )

    assert (
        board.num_clear_positions()
        - args.snake_size
        - num_initial_apples
        - num_spawn_apples
        > 0
    ), instance_name

    snake = board.generate_snake(args.snake_size)
    apples = board.generate_apples(num_initial_apples)
    spawn_apples = board.generate_apples(num_spawn_apples, "B")

    # Not repeated
    assert len(apples) == len(set(apples))
    assert len(spawn_apples) == len(set(spawn_apples))
    assert not set(apples) & set(spawn_apples)

    assert len(apples) == num_initial_apples
    assert len(spawn_apples) == num_spawn_apples, instance_name
    assert len(snake) == args.snake_size + 1

    if args.output:
        sys.stdout = open(f"{args.output}/{instance_name}.pddl", "w")
    if args.output_type == "raw":
        print_solver(instance_name, board, snake, apples, spawn_apples)
    elif args.output_type == "readable":
        print_readable(instance_name, board, snake, apples, spawn_apples)
    elif args.output_type == "pddl":
        print_pddl(instance_name, board, snake, apples, spawn_apples)
    elif args.output_type == "draw":
        print(board)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import random
import itertools

# Generate all combinations from empty to each board.
# Usage ./generate-ipc-instances.py | bash


def get_slack(board):
    if board.startswith("empire"):
        return 1
    elif board.startswith("tower") and not "4" in board:
        return 2
    return 0


def get_height(board):
    if board.startswith("empire"):
        return 1
    elif board.startswith("tower") and not "4" in board:
        return 2
    return 0


random.seed(10)
# Map from board name to slack and max height reduction
# all_boards = {'eiffel.txt' : (2),
#               'empire.txt' : (6),
#               'giza.txt' : (6),
#               'taj_mahal.txt' : (5),
#               'tower.txt' : (7),
#               'tower2.txt': (7) ,
#               'tower4.txt': (7),
#               'wall5x5.txt' : (6)}

all_boards = {
    "eiffel.txt": range(0, 3),
    "empire.txt": range(2, 7),
}


scenarios = [
    ("empty.txt", goal, 0, gh)
    for (goal, range_gh) in all_boards.items()
    for gh in range_gh
]


# + \
# [(init, goal, gh, gh) for ((init, (max_gh1)), (goal, (max_gh2))) in random.sample(sorted(set(itertools.permutations(all_boards.items(), 2))), 10)
#  for gh in range (0, min(max_gh1, max_gh2))
#  ]


for sc in scenarios:
    print(
        "./generate.py boards/{} boards/{} pddlfile --reduce_height_init {} --reduce_height_goal {} --ensure_plan --store_plan".format(
            *sc
        )
    )



""" File:        generator_cmd_line.py
    Author:      Nathan Robinson
    Contact:     nathan.m.robinson@gmail.com
    Date:        2013-11-13

    Desctiption: This file uses cmd_line.py to make a command line interface
                 for the cave diving PDDL problem generator.

    Lisence:    Copyright (c) Year 2013, Nathan Robinson <nathan.m.robinson@gmail.com>
                                         Christian Muise <christian.muise@gmail.com>
                                         Charles Gretton <charles.gretton@gmail.com>

                Permission to use, copy, modify, and/or distribute this software for any
                purpose with or without fee is hereby granted, provided that the above
                copyright notice and this permission notice appear in all copies.

                THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
                WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
                MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
                SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
                WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
                ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
                IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""

import sys

from cmd_line import (
    ArgProcessor,
    ArgDefinition,
    FlagDefinition,
    range_validator,
    seq_range_validator,
    bool_validator,
    InputException,
)


def print_citation_info(arg_processor):
    print("When citing this program please cite the following publication:")
    print("TBA")
    sys.exit(0)


def process_args():
    arg_processor = ArgProcessor()

    arg_processor.add_program_arg(
        "-domain_file",
        ArgDefinition(
            "domain_file_name", False, None, None, None, "The output PDDL domain file"
        ),
    )

    arg_processor.add_program_arg(
        "-problem_file",
        ArgDefinition(
            "problem_file_name", True, None, None, None, "The output PDDL problem file"
        ),
    )

    arg_processor.add_program_arg(
        "-problem_name",
        ArgDefinition(
            "problem_name", True, None, None, None, "The name of the problem instance"
        ),
    )

    arg_processor.add_program_arg(
        "-cave_branches",
        ArgDefinition(
            "cave_branches",
            False,
            seq_range_validator,
            [":", 1, None, int, 2, None, "Error - invalid cave branches:"],
            [3],
            "Depths of the cave branches (x0:...:xk)",
        ),
    )

    arg_processor.add_program_arg(
        "-objectives",
        ArgDefinition(
            "objectives",
            False,
            seq_range_validator,
            [":", 1, None, int, 1, None, "Error - photo objectives:"],
            [3],
            "The depths of the photo objectives (x0:...:xk)",
        ),
    )

    arg_processor.add_program_arg(
        "-num_tank_adjustment",
        ArgDefinition(
            "num_tank_adjustment",
            False,
            range_validator,
            [int, None, None, False, "Error - invalid number of tanks adjustment:"],
            0,
            "How many more/fewer tanks than required to use",
        ),
    )

    arg_processor.add_program_arg(
        "-num_diver_adjustment",
        ArgDefinition(
            "num_diver_adjustment",
            False,
            range_validator,
            [int, None, None, False, "Error - invalid number of divers adjustment:"],
            0,
            "How many more/fewer divers than required to use",
        ),
    )

    arg_processor.add_program_arg(
        "-neg_link_prob",
        ArgDefinition(
            "neg_link_prob",
            False,
            range_validator,
            [float, 0, 1, False, "Error - invalid negative link probability:"],
            0.5,
            "The probability of a negative diver relationship link",
        ),
    )

    arg_processor.add_program_arg(
        "-neg_cycle_length",
        ArgDefinition(
            "neg_cycle_length",
            False,
            range_validator,
            [float, 0, 1, True, "Error - invalid negative cycle length:"],
            None,
            "The frac of divers in a negative cycle (None=no cycle)",
        ),
    )

    arg_processor.add_program_arg(
        "-minimum_hiring_cost",
        ArgDefinition(
            "minimum_hiring_cost",
            False,
            range_validator,
            [int, 0, None, False, "Error - invalid minimum hiring cost:"],
            10,
            "The minimum cost of any hiring action",
        ),
    )

    arg_processor.add_program_arg(
        "-maximum_hiring_cost",
        ArgDefinition(
            "maximum_hiring_cost",
            False,
            range_validator,
            [int, 0, None, False, "Error - invalid maximum hiring cost "],
            100,
            "The maximum cost of any hiring action",
        ),
    )

    arg_processor.add_program_arg(
        "-perturb_hiring_costs",
        ArgDefinition(
            "perturb_hiring_costs",
            False,
            range_validator,
            [float, 0, 1, False, "Error - invalid perturb hiring cost fraction:"],
            0,
            "How much hiring costs should be perturbed",
        ),
    )

    arg_processor.add_program_arg(
        "-other_action_cost",
        ArgDefinition(
            "other_action_cost",
            False,
            range_validator,
            [float, 0, None, False, "Error - invalid other hiring cost:"],
            1,
            "The cost of non-hiring actions",
        ),
    )

    arg_processor.add_program_arg(
        "-strips",
        ArgDefinition(
            "strips",
            False,
            bool_validator,
            ["Error - invalid STRIPS setting:"],
            False,
            "If the domain should be compiled to STRIPS",
        ),
    )

    arg_processor.add_program_arg(
        "-order_tanks",
        ArgDefinition(
            "order_tanks",
            False,
            bool_validator,
            ["Error - invalid order tanks setting:"],
            True,
            "Force an order of the preparation of the tanks",
        ),
    )

    arg_processor.add_program_arg(
        "-seed",
        ArgDefinition(
            "seed",
            False,
            range_validator,
            [int, None, None, True, "Error - invalid seed: "],
            None,
            "Random seed (None for system time)",
        ),
    )

    arg_processor.add_program_flag(
        "--cite",
        FlagDefinition("cite", print_citation_info, "Display citation information"),
    )

    try:
        arg_processor.parse_args()

        if len(arg_processor.objectives) > len(arg_processor.cave_branches):
            raise InputException(
                "Not enough cave branches for number of photo targets."
            )

        if arg_processor.maximum_hiring_cost < arg_processor.minimum_hiring_cost:
            raise InputException("Max hiring cost is less than max.")

    except InputException as e:
        print(e.message)
        print("Use --help flag to display usage information.")
        sys.exit(1)
    return arg_processor



#!/usr/bin/env python
#
# Author: Tomas de la Rosa
# See MIT License attached
#
import sys
import random


def get_tile(row, col):
    return "tile_" + str(row) + "-" + str(col)


# There is an extra row below to place robots at the end


def get_objects():
    #   TILES
    str_tiles = ""
    for row in range(num_rows + 1):
        for column in range(num_columns):
            str_tiles = str_tiles + get_tile(row, column + 1) + " "
        if row != num_rows:
            str_tiles += "\n           "
    str_tiles += "- tile\n"

    # ROBOTS
    str_robots = "           "
    for robot in range(num_robots):
        str_robots += "robot" + str(robot + 1) + " "
    str_robots += "- robot\n"

    # COLORS
    str_colors = "           white black - color\n"
    return str_tiles + str_robots + str_colors


def get_init():
    robot_locations = []
    str_init = "\n"
    if mode_flag != "time":
        str_init += "   (= (total-cost) 0)\n"

    used_cols = list(range(1, num_columns + 1))

    for robot in range(num_robots):
        posx = random.randint(0, num_rows)
        posy = random.choice(used_cols)
        robot_pos = get_tile(posx, posy)
        robot_locations += [robot_pos]
        used_cols.remove(posy)

        str_init += "   (robot-at robot" + str(robot + 1) + " " + robot_pos + ")\n"
        if robot % 2 == 0:
            robot_color = "white"
        else:
            robot_color = "black"
        str_init += "   (robot-has robot" + str(robot + 1) + " " + robot_color + ")\n"

    str_init += "   (available-color white)\n"
    str_init += "   (available-color black)\n"

    for row in range(num_rows + 1):
        for column in range(num_columns):
            if get_tile(row, column + 1) not in robot_locations:
                str_init += "   (clear " + get_tile(row, column + 1) + ")\n"

    for row in range(num_rows):
        for column in range(num_columns):
            str_init += (
                "   (up "
                + get_tile(row + 1, column + 1)
                + " "
                + get_tile(row, column + 1)
                + ")\n"
            )
    for row in range(num_rows):
        for column in range(num_columns):
            str_init += (
                "   (down "
                + get_tile(row, column + 1)
                + " "
                + get_tile(row + 1, column + 1)
                + ")\n"
            )
    for row in range(num_rows + 1):
        for column in range(num_columns - 1):
            str_init += (
                "   (right "
                + get_tile(row, column + 2)
                + " "
                + get_tile(row, column + 1)
                + ")\n"
            )
    for row in range(num_rows + 1):
        for column in range(num_columns - 1):
            str_init += (
                "   (left "
                + get_tile(row, column + 1)
                + " "
                + get_tile(row, column + 2)
                + ")\n"
            )

    return str_init


def get_goals():
    str_goals = "(and\n"
    for row in range(num_rows):
        for column in range(num_columns):
            if (row + column) % 2 == 0:
                str_goals += (
                    "    (painted " + get_tile(row + 1, column + 1) + " white)\n"
                )
            else:
                str_goals += (
                    "    (painted " + get_tile(row + 1, column + 1) + " black)\n"
                )
    return str_goals + ")"


# Reading the command line arguments
try:
    name = sys.argv[1]
    num_rows = int(sys.argv[2])
    num_columns = int(sys.argv[3])
    num_robots = int(sys.argv[4])
    mode_flag = sys.argv[5]

    if len(sys.argv) == 7:
        seed = int(sys.argv[6])
    else:
        seed = None

    random.seed(seed)

except BaseException:
    print(
        "Usage: "
        + sys.argv[0]
        + " <name> <num_rows> <num_columns> <num_robots> <mode_flag(seq|time)> [<seed>]"
    )
    sys.exit(1)

print("(define (problem " + name + ")")
print(" (:domain floor-tile)")
print(" (:objects " + get_objects() + ")")
print(" (:init " + get_init() + ")")
print(" (:goal " + get_goals() + ")")

if mode_flag == "time":
    print(" (:metric minimize (total-time))")
else:
    print(" (:metric minimize (total-cost))")

print(")")

sys.exit(0)



#!/usr/bin/env python
import sys
import random

NUM_MATCHES = 1

# *****************#
# Functions declarations
# *****************#


def get_objects(fuses):
    str_objects = "\n   "
    for i in range(0, NUM_MATCHES):
        str_objects = str_objects + " match" + str(i)
    str_objects = str_objects + " - match"

    str_objects = str_objects + "\n   "

    for i in range(0, fuses):
        str_objects = str_objects + " fuse" + str(i)
    str_objects = str_objects + " - fuse\n"

    return str_objects


# *****************#


def get_init():
    str_init = "\n"
    str_init = str_init + "  (handfree)\n"

    for i in range(0, NUM_MATCHES):
        str_init = str_init + "  (unused match" + str(i) + ")\n"

    return str_init


# *****************#


def get_goals():
    str_goal = ""
    str_goal = str_goal + "\n  (and\n"
    for i in range(0, NUM_FUSES):
        str_goal = str_goal + "     (mended fuse" + str(i) + ")\n"

    str_goal = str_goal + ")"
    return str_goal


# *****************#
# MAIN
# *****************#
# Reading the command line arguments
import argparse
parser = argparse.ArgumentParser()
parser.add_argument("matches", type=int, help="number of matches")
parser.add_argument("--seed", type=int, help="random seed")
args = parser.parse_args()
NUM_MATCHES = args.matches
if args.seed is not None:
    random.seed(args.seed)


NUM_FUSES = NUM_MATCHES * 2

print("(define (problem prob)")
print(" (:domain matchcellar)")
print(" (:objects " + get_objects(NUM_FUSES) + ")")
print(" (:init " + get_init() + ")")
print(" (:goal" + get_goals() + ")")
print(" (:metric minimize (total-time))")
print(")")

sys.exit(0)



import logging
import os
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property
from typing import List, TypeVar, Set, Dict, Any

from unified_planning.io import PDDLReader, PDDLWriter
from unified_planning.model import Problem, Fluent, UPState
from unified_planning.plans import ActionInstance
from unified_planning.shortcuts import SequentialSimulator, BoolType, FALSE, TRUE

from tarski.io import PDDLReader as tarskiPDDLReader
from tarski.grounding import LPGroundingStrategy

from amlgym.modeling.env import Env

ObservationType = TypeVar('ObservationType')


@dataclass
class UPEnv(Env):
    """
    A Gymnasium environment which simulates a PDDL problem
    through the unified-planning sequential simulator.
    """

    # The PDDL problem in unified-planning
    problem: Problem

    # The environment simulation engine
    _simulator: SequentialSimulator

    def __init__(self,
                 domain_path: str,
                 problem_path: str) -> None:
        """
        Set environment state and seed through :meth:`reset` for reproducibility
        """

        # Parse the PDDL environment in unified-planning
        self.problem = PDDLReader().parse_problem(domain_path, problem_path)

        # Instantiate the environment simulation engine
        self._simulator = SequentialSimulator(self.problem)

        # Create a fictitious state with all negated literals
        all_neg_fluents = {f: FALSE() for f, v in self.problem.initial_values.items()}
        UPState.MAX_ANCESTORS = None
        self.all_neg_state = UPState(all_neg_fluents, self.problem)

        # Initialize actions grounder with tarski
        _tmp_problem = PDDLReader().parse_problem(domain_path, problem_path)
        # Add a dummy fluent to show `preconditions:` and `effects:` sections in the PDDL file
        dummy_fluent = Fluent('dummy', BoolType())
        if dummy_fluent not in _tmp_problem.fluents:
            _tmp_problem.add_fluent(dummy_fluent)
        _tmp_problem.set_initial_value(dummy_fluent, True)
        # Rebuild actions with no preconditions/effects
        for action in _tmp_problem.actions:
            action.clear_preconditions()
            action.clear_effects()
            # ensure `preconditions:` and `effects:` sections in the PDDL file
            action.add_precondition(dummy_fluent)
            action.add_effect(dummy_fluent, True)
        # Remove problem goal to avoid tarski reachability issues
        _tmp_problem.clear_goals()
        tmp_domain_path = 'tmp_domain.pddl'
        tmp_problem_path = 'tmp_problem.pddl'
        PDDLWriter(_tmp_problem).write_domain(tmp_domain_path)
        PDDLWriter(_tmp_problem).write_problem(tmp_problem_path)
        reader = tarskiPDDLReader(raise_on_error=True)
        reader.parse_domain(tmp_domain_path)
        reader.parse_instance(tmp_problem_path)
        os.remove(tmp_domain_path)
        os.remove(tmp_problem_path)
        self._grounder = LPGroundingStrategy(reader.problem)

    @cached_property
    # def ground_actions(self) -> List[ActionInstance]:
    def ground_actions(self) -> Dict[str, Any]:
        """
        Return a list of all ground actions for the current environment.
        :return: ground actions list
        """
        logging.debug("Grounding actions with tarski...")
        ground_actions = self._grounder.ground_actions()
        return ground_actions

    def _str_to_action(self, action_label: str) -> ActionInstance:
        """
        Get UP problem ground action from the action label
        :param action_label: action label string
        :return: unified planning action instance
        """
        action_split = action_label.strip()[1:-1].split()
        op_name = action_split[0]
        if len(action_split) > 1:
            obj_names = [o.strip() for o in action_split[1:]]
        else:
            obj_names = []
        up_op = self.problem.action(op_name)
        up_objs = [self.problem.object(o) for o in obj_names]
        return ActionInstance(up_op, up_objs)

    def apply(self, state, action):
        """
        Return the state :math:`s'` reached after executing action :math:`a`
        in state :math:`s`.
        :param state: current state :math:`s`
        :param action: action :math:`a` to be executed
        :return: future state :math:`s'`
        """
        if isinstance(action, str):
            action = self._str_to_action(action)

        if isinstance(state, Set) or isinstance(state, List):

            pos_literals = {l for l in state if not l.startswith('(not ')}

            prob_state_fluents = dict()
            for f in pos_literals:
                f_split = f[1:-1].split()
                f_name = f_split[0]
                f_objs = []
                if len(f_split) > 1:
                    f_objs = f_split[1:]

                prob_f = self.problem.fluent(f_name)
                prob_args = [self.problem.object(o) for o in f_objs]
                prob_state_fluents[prob_f(*prob_args)] = TRUE()

            state = self.all_neg_state.make_child(prob_state_fluents)

        next_state = self._simulator.apply(state, action)

        literals = set()
        for l, v in next_state._values.items():
            l_name = l.fluent().name
            l_objs = [str(o) for o in l.args]

            if len(l_objs) == 0:
                l_formatted = f"({l_name})"
            else:
                l_formatted = f"({l_name} {' '.join(l_objs)})"

            if v.is_true():
                literals.add(l_formatted)
            else:
                literals.add(f"(not {l_formatted})")

        return literals

    def applicable_actions(self, state) -> Dict[str, Set[str]]:

        # cached = self.cache_app_actions.get(frozenset(state), None)
        # if cached is not None:
        #     return cached

        if isinstance(state, Set) or isinstance(state, List):

            pos_literals = {l for l in state if not l.startswith('(not ')}

            prob_state_fluents = dict()
            for f in pos_literals:
                f_split = f[1:-1].split()
                f_name = f_split[0]
                f_objs = []
                if len(f_split) > 1:
                    f_objs = f_split[1:]

                prob_f = self.problem.fluent(f_name)
                prob_args = [self.problem.object(o) for o in f_objs]
                prob_state_fluents[prob_f(*prob_args)] = TRUE()

            state = self.all_neg_state.make_child(prob_state_fluents)

        applicable_actions = defaultdict(set)
        for op_name, param_combos in self.ground_actions.items():
            for args in param_combos:
                if self._simulator._is_applicable(state,
                                                  self.problem.action(op_name),
                                                  [self.problem.object(o.lower())
                                                   for o in args]):
                    if len(args) > 0:
                        action_label = f"({op_name} {' '.join(args)})"
                    else:
                        action_label = f"({op_name})"

                    applicable_actions[op_name].add(action_label)

        # self.cache_app_actions[frozenset(state)] = applicable_actions
        return applicable_actions



#!/usr/bin/env python

import argparse
import random

from solve import solve


def generate_readable(
    instance_name, grid_size_x, grid_size_y, max_height, robots, depots, i_grid, g_grid
):
    assert len(i_grid) == grid_size_y
    assert len(i_grid[0]) == grid_size_x
    lines = [instance_name]
    lines += ["Initial state:"]
    for y in range(grid_size_y):
        cells = []
        for x in range(grid_size_x):
            r = "R" if (x, y) in robots else " "
            h = str(i_grid[y][x])
            d = "D" if (x, y) in depots else " "
            cells.append(r + h + d)
        lines += [" ".join(cells)]
    lines += ["Goal state:"]
    for y in range(grid_size_y):
        cells = []
        for x in range(grid_size_x):
            if g_grid[y][x] is None:
                cells += [" * "]
            else:
                cells += [" %s " % g_grid[y][x]]
        lines += [" ".join(cells)]
    lines += ["Maximal height: %d" % max_height]
    return lines


def generate_for_pddl(
    instance_name,
    grid_size_x,
    grid_size_y,
    max_height,
    robots,
    depots,
    i_grid,
    g_grid,
    MULTI_ROBOT,
):
    comments = generate_readable(
        instance_name,
        grid_size_x,
        grid_size_y,
        max_height,
        robots,
        depots,
        i_grid,
        g_grid,
    )

    objects = []
    objects += ["n%d - numb" % (n) for n in range(max_height + 1)]
    objects += [
        "pos-%d-%d - position" % (x, y)
        for x in range(grid_size_x)
        for y in range(grid_size_y)
    ]

    facts = []
    facts += [
        "(height pos-%d-%d n%d)" % (x, y, i_grid[y][x])
        for x in range(grid_size_x)
        for y in range(grid_size_y)
    ]

    if MULTI_ROBOT:
        objects += ["r-%s - robot" % robot_id for robot_id in range(len(robots))]
        facts += ["(occupied pos-%d-%d)" % (x, y) for (x, y) in robots]
        facts += [
            "(at r-%d pos-%d-%d)" % (robot_id, x, y)
            for robot_id, (x, y) in enumerate(robots)
        ]
    else:
        assert len(robots) == 1
        facts += ["(at pos-%d-%d)" % (x, y) for (x, y) in robots]

    # Static predicates
    facts += ["(SUCC n%d n%d)" % (n + 1, n) for n in range(max_height)]
    facts += [
        "(NEIGHBOR pos-%d-%d pos-%d-%d)" % (x, y, x + dx, y + dy)
        for x in range(grid_size_x)
        for y in range(grid_size_y)
        for dx, dy in [(-1, 0), (1, 0), (0, -1), (0, 1)]
        if 0 <= x + dx < grid_size_x and 0 <= y + dy < grid_size_y
    ]
    facts += ["(IS-DEPOT pos-%d-%d)" % (x, y) for (x, y) in depots]

    goals = [
        "(height pos-%d-%d n%d)" % (x, y, g_grid[y][x])
        for x in range(grid_size_x)
        for y in range(grid_size_y)
        if g_grid[y][x] is not None
    ]
    goals += ["(not (has-block))"]

    lines = []
    lines += ["(define (problem %s)" % instance_name]
    lines += ["(:domain termes)"]
    lines += ["; %s" % c for c in comments]
    lines += ["(:objects"]
    lines += ["    %s" % o for o in objects]
    lines += [")", "(:init"]
    lines += ["    %s" % f for f in facts]
    lines += [")", "(:goal (and"]
    lines += ["    %s" % g for g in goals]
    lines += ["))"]
    lines += [")"]
    return lines


def remove_slack(grid):
    num_rows = len(grid)
    num_columns = len(grid[0])
    non_null_rows = sorted(
        {y for x in range(num_columns) for y in range(num_rows) if grid[y][x] > 0}
    )

    if not non_null_rows:
        return [[0]]

    grid = grid[non_null_rows[0] : non_null_rows[-1] + 1]

    num_rows = len(grid)
    num_columns = len(grid[0])

    non_null_columns = sorted(
        {x for x in range(num_columns) for y in range(num_rows) if grid[y][x] > 0}
    )

    grid = list(map(lambda x: x[non_null_columns[0] : non_null_columns[-1] + 1], grid))

    return grid


def get_value(grid, x, y, dx, dy):
    rx = x - dx
    ry = y - dy

    if ry < 0 or ry >= len(grid) or rx < 0 or rx >= len(grid[ry]):
        return 0

    return grid[ry][rx]


def parse():
    parser = argparse.ArgumentParser()

    parser.add_argument("init", type=argparse.FileType("r"))
    parser.add_argument("goal", type=argparse.FileType("r"))
    parser.add_argument("--reduce_height_init", type=int, default=0)
    parser.add_argument("--reduce_height_goal", type=int, default=0)
    parser.add_argument("--grid_slack", type=int, default=0)
    parser.add_argument("--multi_robot", action="store_true")
    parser.add_argument("--ensure_plan", action="store_true")
    parser.add_argument("--store_plan", action="store_true")
    parser.add_argument("--dont_remove_slack", action="store_true")

    # parser.add_argument("random_seed", type=int)
    parser.add_argument("output_type", choices=["pddl", "pddlfile", "readable"])
    return parser.parse_args()


def place_depots(grid_size_x, grid_size_y, init_grid, goal_grid):
    candidate_cells = [
        (y, abs(x - grid_size_x / 2), x, y)
        for x in range(grid_size_x)
        for y in range(grid_size_y)
        if init_grid[y][x] == 0 and goal_grid[y][x] == 0
    ]
    if not candidate_cells:
        return None
    selected = min(candidate_cells)
    return [(selected[2], selected[3])]


def main():
    args = parse()

    i_grid = [
        list(map(lambda x: max(0, int(x) - args.reduce_height_init), l.strip().split()))
        for l in args.init.readlines()
    ]
    g_grid = [
        list(map(lambda x: max(0, int(x) - args.reduce_height_goal), l.strip().split()))
        for l in args.goal.readlines()
    ]

    if not args.dont_remove_slack:
        i_grid = remove_slack(i_grid)
        g_grid = remove_slack(g_grid)

    grid_size_x = max(len(i_grid[0]), len(g_grid[0])) + args.grid_slack
    grid_size_y = max(len(i_grid), len(g_grid)) + args.grid_slack

    increase_x = grid_size_x < grid_size_y
    depots = None
    while not depots:
        dxinit = (grid_size_x - len(i_grid[0])) // 2 + (
            (grid_size_x - len(i_grid[0])) % 2
        )
        dyinit = (grid_size_y - len(i_grid)) // 2 + ((grid_size_y - len(i_grid)) % 2)

        dxgoal = (grid_size_x - len(g_grid[0])) // 2 + (
            (grid_size_x - len(g_grid[0])) % 2
        )
        dygoal = (grid_size_y - len(g_grid)) // 2 + ((grid_size_y - len(g_grid)) % 2)

        init_grid = [
            [get_value(i_grid, x, y, dxinit, dyinit) for x in range(0, grid_size_x)]
            for y in range(0, grid_size_y)
        ]
        goal_grid = [
            [get_value(g_grid, x, y, dxgoal, dygoal) for x in range(0, grid_size_x)]
            for y in range(0, grid_size_y)
        ]

        depots = place_depots(grid_size_x, grid_size_y, init_grid, goal_grid)
        if not depots:
            if increase_x:
                grid_size_x += 1
            else:
                grid_size_y += 1
            increase_x = not increase_x

    robots = [depots[0]]  # A single robot, placed in the first depot

    max_height = max(list(map(max, init_grid)) + list(map(max, goal_grid)))

    problem_size = grid_size_x * grid_size_y * max_height

    if args.ensure_plan:
        plan = solve(
            grid_size_x, grid_size_y, max_height, robots, depots, init_grid, goal_grid
        )
        while not plan:
            if increase_x:
                grid_size_x += 1
            else:
                grid_size_y += 1
            increase_x = not increase_x

            dxinit = (grid_size_x - len(i_grid[0])) // 2 + (
                (grid_size_x - len(i_grid[0])) % 2
            )
            dyinit = (grid_size_y - len(i_grid)) // 2 + (
                (grid_size_y - len(i_grid)) % 2
            )

            dxgoal = (grid_size_x - len(g_grid[0])) // 2 + (
                (grid_size_x - len(g_grid[0])) % 2
            )
            dygoal = (grid_size_y - len(g_grid)) // 2 + (
                (grid_size_y - len(g_grid)) % 2
            )

            init_grid = [
                [get_value(i_grid, x, y, dxinit, dyinit) for x in range(0, grid_size_x)]
                for y in range(0, grid_size_y)
            ]
            goal_grid = [
                [get_value(g_grid, x, y, dxgoal, dygoal) for x in range(0, grid_size_x)]
                for y in range(0, grid_size_y)
            ]

            depots = place_depots(grid_size_x, grid_size_y, init_grid, goal_grid)
            robots = depots

            plan = solve(
                grid_size_x,
                grid_size_y,
                max_height,
                robots,
                depots,
                init_grid,
                goal_grid,
            )

    instance_name_list = ["termes"]

    if args.ensure_plan:
        instance_name_list.append(f"{len(plan):05d}")

    instance_name_list += [
        f"{problem_size:04d}",
        f"{grid_size_x}x{grid_size_y}x{max_height}",
    ]

    if not args.init.name.split("/")[-1].startswith("empty"):
        instance_name_list += [args.init.name.split("/")[-1].replace(".txt", "")]

    instance_name_list += [args.goal.name.split("/")[-1].replace(".txt", "")]

    instance_name = "-".join(instance_name_list)

    if args.output_type.startswith("pddl"):
        lines = generate_for_pddl(
            instance_name,
            grid_size_x,
            grid_size_y,
            max_height,
            robots,
            depots,
            init_grid,
            goal_grid,
            args.multi_robot,
        )
        if args.output_type == "pddl":
            print("\n".join(lines))
        else:
            f = open(f"instances/p-{instance_name}.pddl", "w")
            f.write("\n".join(lines))
            f.close()
    elif args.output_type == "readable":
        lines = generate_readable(
            instance_name,
            grid_size_x,
            grid_size_y,
            max_height,
            robots,
            depots,
            init_grid,
            goal_grid,
        )

        print("\n".join(lines))

    if args.store_plan:
        if not plan:
            plan = solve(
                grid_size_x,
                grid_size_y,
                max_height,
                robots,
                depots,
                init_grid,
                goal_grid,
            )
        if plan:
            f = open(f"plans/p-{instance_name}.pddl", "w")
            f.write("\n".join(plan))
            f.close()


if __name__ == "__main__":
    main()



#! /usr/bin/env python


import math
import random


USAGE = """usage:
    %s MIN_SIZE MAX_SIZE STEP_SIZE MIN_ID MAX_ID DENSITY
where
    MIN_SIZE   the minimum problem size (number of products and orders)
    MAX_SIZE   the maximum problem size (number of products and orders)
    STEP_SIZE  the increment on problem size
    MIN_ID     the start number of problems of the same size
    MAX_ID     the end number of problems of the same size
    DENSITY    the approximate percentage of products included in each order"""


def main(args):
    if len(args) != 7:
        raise SystemExit(USAGE % args[0])
    min_size, max_size, step_size, min_id, max_id, density = list(map(int, args[1:]))
    for size in range(min_size, max_size + 1, step_size):
        for counter in range(min_id, max_id + 1):
            filename = "p%d_%d.txt" % (size, counter)
            generate_problem(filename, size, size, density)


def include_product_for_order(product, order, num_products, density):
    # This is based on Ioannis's code. It is designed so that products
    # are more likely to be included in orders if the distance between
    # the product and order id (before the reshuffling that happens at
    # the end) is small.
    #
    # Compared to an independent identically distributed choice for
    # each product/order pair, this should tend to favour instances
    # with small (= few stacks) solutions since it adds some locality
    # to the problem: the matrix entries (before shuffling) will tend
    # to cluster around the diagonal.

    d = abs(product - order)
    s = num_products / 6.0
    # The next line is the density function of the normal distribution
    # with mu = 0 and sigma = s for x = d.
    p = (1 / (s * (2 * math.pi) ** 0.5)) * math.exp(-(d * d) / (2 * s * s))
    # The next line is rather magical to me. If there is an
    # explanation for what this formula signifies, I don't know it.
    chance = p * density * num_products / (2.7 * (num_products / 5.0) ** 0.5)
    return random.randrange(100) <= chance


def generate_problem(filename, num_products, num_orders, density):
    products = list(range(num_products))
    orders = list(range(num_orders))
    matrix = [[0 for product in products] for order in orders]
    for order in orders:
        for product in products:
            if include_product_for_order(product, order, num_products, density):
                matrix[order][product] = 1
    for order in orders:
        if all(matrix[order][product] == 0 for product in products):
            matrix[order][random.choice(products)] = 1
    for product in products:
        if all(matrix[order][product] == 0 for order in orders):
            matrix[random.choice(orders)][product] = 1
    weights = [random.randrange(10, 100 + 1, 10) for product in products]

    random.shuffle(products)
    # Ioannis's original code, used for generating the competition
    # instances, only shuffles the orders but not the products, which
    # means that planners that process the products in numerical order
    # (or whatever) are somewhat biased. (Depending on the planner
    # and/or track, I guess this can be either positive or negative.)
    random.shuffle(orders)

    with open(filename, "w") as outfile:
        print(num_orders, num_products, file=outfile)
        for order in orders:
            row = [matrix[order][product] for product in products]
            print(" ".join(map(str, row)), file=outfile)
        print(" ".join(map(str, weights)), file=outfile)


if __name__ == "__main__":
    import sys

    main(sys.argv)



r"""
Problem solving metrics assess a domain model ability of solving problems.
The *solving ratio* evaluates the model ability of producing plans that are
applicable in the environment and achieve the given goal.
The *false plans ratio* measures the ratio of produced plans that are either not
applicable in the environment or do not achieve the given goal.
Problem solving metrics are defined with respect to a set of problems and a planner.
"""
import contextlib
import os
import warnings

from alive_progress import alive_bar
from unified_planning.engines import PlanGenerationResultStatus, ValidationResultStatus
from unified_planning.io import PDDLReader, PDDLWriter
from unified_planning.shortcuts import PlanValidator, OneshotPlanner
from typing import Dict, List


def problem_solving(model_learn_path: str,
                    model_ref_path: str,
                    problem_paths: List[str],
                    timeout=60,
                    show_progress: bool = True) -> Dict[str, float]:
    """
    Solve the given problems using the model to be evaluated and return:
     (i) the solving plans ratio in the environment defined by the reference model
     (ii) the false plans ratio in the environment defined by the reference model
     (iii) the ratio of problems deemed unsolvable
     (iv) the ratio of problems where the planning process timed out
     (v) the ratio of problems that raised syntax errors during parsing in unified planning

    :param model_learn_path: learned model path
    :param model_ref_path: reference model path
    :param problem_paths: list of problem paths
    :param timeout: planner timeout

    :return: solving and false plans ratios, deemed unsolvable problems, timed out planning processes
    """
    reader = PDDLReader()
    DOWNWARD_SEARCH_CFG = 'let(hff,ff(),let(hcea,cea(),lazy_greedy([hff,hcea],preferred=[hff,hcea])))'
    HEUR_PLANNER_CFG = {
        'name': 'fast-downward',
        'params': dict(
            fast_downward_search_config=DOWNWARD_SEARCH_CFG,
            fast_downward_search_time_limit=f"{timeout}s"
        )}

    solving = 0
    false_plans = 0
    unsolvable = 0
    timed_out = 0
    syntax_errors = 0

    # Solve the problem with the learned model
    bar = alive_bar(len(problem_paths),
                    title=f'Evaluating problem solving...',
                    length=20) if show_progress else contextlib.nullcontext()
    with bar as bar:
        for problem_path in problem_paths:

            try:
                problem = reader.parse_problem(model_learn_path, problem_path)
            except SyntaxError as err:
                warnings.warn(
                    f"Failed to parse problem '{problem_path}' with domain "
                    f"{model_learn_path}: {err}",
                    stacklevel=2
                )
                syntax_errors += 1
                continue

            with contextlib.redirect_stdout(open(os.devnull, 'w')):
                with OneshotPlanner(
                        problem_kind=problem.kind,
                        **HEUR_PLANNER_CFG
                ) as planner:
                    result = planner.solve(problem, timeout=timeout)
                    plan = result.plan

            if plan is not None:  # neither solving_plan nor false_plan
                # Parse problem
                problem_ref = reader.parse_problem(model_ref_path, problem_path)

                PDDLWriter(problem_ref).write_plan(plan, 'tmp')

                if validate_plan(model_ref_path, problem_path, 'tmp'):
                    solving += 1
                else:
                    false_plans += 1

                os.remove('tmp')
            else:
                if result.status == PlanGenerationResultStatus.TIMEOUT:
                    timed_out += 1
                elif result.status in [PlanGenerationResultStatus.UNSOLVABLE_PROVEN,
                                       PlanGenerationResultStatus.UNSOLVABLE_INCOMPLETELY]:
                    unsolvable += 1

            if show_progress:
                bar()

    return {
        'solving_ratio': solving / len(problem_paths),
        'false_plans_ratio': false_plans / len(problem_paths),
        'unsolvable_ratio': unsolvable / len(problem_paths),
        'timed_out': timed_out / len(problem_paths),
        'syntax_errors': syntax_errors / len(problem_paths)
    }


def validate_plan(model_path: str,
                  problem_path: str,
                  plan_path: str,) -> float:
    """
    Validate the solution plan to the given problem in the environment defined by the given model.
    The definition of valid plan follows the one provided in "PDDL2. 1: An extension to PDDL for
    expressing temporal planning domains", M. Fox and D. Long, JAIR 2003.

    :param model_path: path to the PDDL model defining the environment.
    :param problem_path: path to the PDDL problem to be solved.
    :param plan_path: path to the solution plan for the PDDL problem.

    :return: true if the plan is valid, otherwise false.
    """
    reader = PDDLReader()
    # Parse problem
    problem = reader.parse_problem(model_path, problem_path)
    # Parse plan
    plan = reader.parse_plan(problem, plan_path)

    with PlanValidator() as validator:
        result = validator.validate(problem, plan)

    return result.status == ValidationResultStatus.VALID



#! /usr/bin/env python

import itertools


def split_into_blocks(lines):
    block = []
    for line in lines:
        line = line.partition(";")[0].rstrip()
        if line:
            block.append(line)
        elif block:
            yield block
            block = []
    if block:
        yield block


def split_descriptions(lines):
    comments = []
    maze = []
    for block in split_into_blocks(lines):
        for line in block:
            if line.startswith(('"', "'")):
                comments.append(line)
                assert not maze
            else:
                maze.append(line)
        if maze:
            yield comments, maze
            maze = []
            comments = []


class Direction:
    def __init__(self, name, dy=0, dx=0):
        self.name = name
        self.dy = dy
        self.dx = dx

    def __call__(self, row, col):
        return (row + self.dy, col + self.dx)


def create_pddl(filename, prob_name, desc, track, hex):
    assert track in ["sequential", "temporal"], track

    comments, maze = desc
    num_rows = len(maze)
    num_cols = max(len(row) for row in maze)
    maze = [row.ljust(num_cols) for row in maze]

    player_counter = itertools.count(1)
    stone_counter = itertools.count(1)

    init = []
    objects = []
    goal = []

    if hex:
        dirs = [
            Direction("dir-northwest", dx=-1, dy=-1),
            Direction("dir-west", dx=-2),
            Direction("dir-southwest", dx=-1, dy=+1),
            Direction("dir-northeast", dx=+1, dy=-1),
            Direction("dir-east", dx=+2),
            Direction("dir-southeast", dx=+1, dy=+1),
        ]
    else:
        dirs = [
            Direction("dir-up", dy=-1),
            Direction("dir-down", dy=+1),
            Direction("dir-left", dx=-1),
            Direction("dir-right", dx=+1),
        ]

    def pos_name(row, col):
        width = len(str(max(num_rows, num_cols)))
        return "pos-%0*d-%0*d" % (width, col + 1, width, row + 1)

    for row in range(num_rows):
        for col in range(num_cols):
            cell = maze[row][col]
            assert cell in " #$.@+*", cell
            if hex and (row + col) % 2:
                # Hexoban uses only every 2nd cell.
                assert cell == " ", (row, col, cell)
                continue
            is_player = cell in "@+"
            is_stone = cell in "$*"
            is_goal = cell in ".*+"
            is_wall = cell in "#"

            pos = pos_name(row, col)
            objects.append("%s - location" % pos)

            if is_goal:
                init.append("(IS-GOAL %s)" % pos)
            else:
                init.append("(IS-NONGOAL %s)" % pos)
            if not is_player and not is_stone and not is_wall:
                init.append("(clear %s)" % pos)
            if is_player:
                player = "player-%02d" % next(player_counter)
                objects.append("%s - player" % player)
                init.append(f"(at {player} {pos})")
            if is_stone:
                stone = "stone-%02d" % next(stone_counter)
                objects.append("%s - stone" % stone)
                init.append(f"(at {stone} {pos})")
                if is_goal:
                    init.append("(at-goal %s)" % stone)
                goal.append("(at-goal %s)" % stone)
            for direction in dirs:
                row2, col2 = direction(row, col)
                if 0 <= col2 < num_cols and 0 <= row2 < num_rows:
                    if not is_wall and maze[row2][col2] != "#":
                        pos2 = pos_name(row2, col2)
                        init.append(f"(MOVE-DIR {pos} {pos2} {direction.name})")
    for direction in dirs:
        objects.append("%s - direction" % direction.name)

    out = open(filename, "w")
    for line in comments:
        print(";; %s" % line.rstrip(), file=out)
    if comments:
        print(";;", file=out)
    for line in maze:
        print(";; %s" % line.rstrip(), file=out)
    print(file=out)
    print("(define (problem %s)" % prob_name, file=out)
    print("  (:domain sokoban-%s)" % track, file=out)
    print("  (:objects", file=out)
    for objdesc in sorted(objects):
        print("    %s" % objdesc, file=out)
    print("  )", file=out)
    print("  (:init", file=out)
    for initdesc in sorted(init):
        print("    %s" % initdesc, file=out)
    if track == "sequential":
        print("    (= (total-cost) 0)", file=out)
    print("  )", file=out)
    print("  (:goal (and", file=out)
    for goaldesc in sorted(goal):
        print("    %s" % goaldesc, file=out)
    print("  ))", file=out)
    if track == "sequential":
        print("  (:metric minimize (total-cost))", file=out)
    else:
        print("  (:metric minimize (total-time))", file=out)
    print(")", file=out)
    out.close()


def translate_suite(suite, hex=False):
    input_file = open("sources/%s.data" % suite)
    for no, desc in enumerate(split_descriptions(input_file)):
        for track in ["sequential", "temporal"]:
            name = "p%03d-%s-%s" % (no + 1, suite, track)
            create_pddl("pddl/%s.pddl" % name, name, desc, track, hex)


if __name__ == "__main__":
    translate_suite("microban")
    translate_suite("multiban")
    translate_suite("hexoban", hex=True)
